package mngr

import (
	"gitlab.smartblocklab.com/smartcontracts/sc-security/actions"
	"gitlab.smartblocklab.com/smartcontracts/nums"
)

// Enumerate chaincode actions
var (
	AddOffer                        = actions.Actions().OfferCC().AddOffer.FunctionName
	GetOffer                        = actions.Actions().OfferCC().GetOffer.FunctionName
	ChangeOfferStatusByOperationId  = actions.Actions().OfferCC().ChangeOfferStatusByOperationId.FunctionName
	GetRequestsByStatus             = actions.Actions().OfferCC().GetRequestsByStatus.FunctionName
	GetUserIncomingRequestsByStatus = actions.Actions().OfferCC().GetUserIncomingRequestsByStatus.FunctionName
	GetUserOutgoingRequestsByStatus = actions.Actions().OfferCC().GetUserOutgoingRequestsByStatus.FunctionName
	GetHistoryUserIncomingRequests  = actions.Actions().OfferCC().GetHistoryUserIncomingRequests.FunctionName
	GetHistoryUserOutgoingRequests  = actions.Actions().OfferCC().GetHistoryUserOutgoingRequests.FunctionName
	IsOfferApproved                 = actions.Actions().OfferCC().IsOfferApproved.FunctionName
	ReplayOffer                     = actions.Actions().OfferCC().ReplayOffer.FunctionName
	GetReplayHistoryOffer          = actions.Actions().OfferCC().GetReplayHistoryOffer.FunctionName
)

// OfferRequest keeps information relevant for user offering request
type OfferRequest struct {
	OperationID   string             `json:"operation_id,omitempty"`
	OutUserInfo   OfferRequestInfo   `json:"out_info,omitempty"`
	InUserInfo    OfferRequestInfo   `json:"in_info,omitempty"`
	RequestStatus OfferRequestStatus `json:"request_status,omitempty"`
	ReplayedID    string             `json:"replayed_id,omitempty"`
}

// InOfferRequest parameter for getQueryResult
type InOfferRequest struct {
	OperationID   string             `json:"operation_id,omitempty"`
	InUserInfo    OfferRequestInfo   `json:"in_info,omitempty"`
	RequestStatus OfferRequestStatus `json:"request_status,omitempty"`
}

// OutOfferRequest parameter for getQueryResult
type OutOfferRequest struct {
	OperationID   string             `json:"operation_id,omitempty"`
	OutUserInfo   OfferRequestInfo   `json:"out_info,omitempty"`
	RequestStatus OfferRequestStatus `json:"request_status,omitempty"`
}

// UsersOffers encapsulates all users offers
type UsersOffers struct {
	UserID string          `json:"user_id,omitempty"`
	Offers []*OfferRequest `json:"user_offers,omitempty"`
}

// OffersList keeps list of offers
type OffersList struct {
	Offers []*OfferRequest `json:"user_offers,omitempty"`
}

// OfferRequestInfo capture offer information relevant to request
type OfferRequestInfo struct {
	UserID       string       `json:"user_id,omitempty"`
	WalletID     string       `json:"wallet_id,omitempty"`
	AccountID    string       `json:"account_id,omitempty"`
	Amount       nums.PlainCurrencyNumber      `json:"amount,omitempty"`
}

// OfferRequestStatus type to represent offer request status
type OfferRequestStatus string

// Enumerate possible statuses for users offers requests
const (
	OfferPending   OfferRequestStatus = "OfferPending"
	OfferAccepted  OfferRequestStatus = "OfferAccepted"
	OfferRejected  OfferRequestStatus = "OfferRejected"
	OfferCanceled  OfferRequestStatus = "OfferCanceled"
	OfferCompleted OfferRequestStatus = "OfferCompleted"
)

type Id struct {
	ID string `json:"id"`
}

type GetUserIncomingRequestsByStatusArg struct {
	UserID             string `json:"user_id"`
	OfferRequestStatus string `json:"offer_request_status"`
}

type GetRequestsByStatusArg struct {
	OfferRequestStatus string `json:"offer_request_status"`
}

type GetUserOutgoingRequestsByStatusArg struct {
	UserID             string `json:"user_id"`
	OfferRequestStatus string `json:"offer_request_status"`
}

type GetHistoryUserOutgoingRequestsArg struct {
	UserID string `json:"user_id"`
}

type GetHistoryUserIncomingRequestsArg struct {
	UserID string `json:"user_id"`
}

type IsOfferApprovedArg struct {
	OperationID string `json:"operation_id"`
}

type ChangeOfferStatusByOperationIdArg struct {
	OperationID        string `json:"operation_id"`
	OfferRequestStatus string `json:"offer_request_status"`
}

//

// AccountInfo ids which uniquely identifies
// information about certain account which has to be validated
type AccountInfo struct {
	UserID    string `json:"user_id,omitempty"`
	WalletID  string `json:"wallet_id,omitempty"`
	AccountID string `json:"account_id,omitempty"`
	Status    Status `json:"status,omitempty"`
}

// WalletsEnvelope struct to encapsulate list of
// all user wallets
type WalletsEnvelope struct {
	UserID  string                `json:"user_id,omitempty"`
	Wallets []*WalletWithAccounts `json:"wallet_list,omitempty"`
}

type WalletWithAccounts struct {
	Wallet   Wallet    `json:"wallet,omitempty"`
	Accounts []Account `json:"accounts,omitempty"`
}

// Account represents system account which includes
// information about account type and account wallets
type Account struct {
	DocType      string       `json:"doc_type,omitempty"`
	AccountID    string       `json:"account_id,omitempty"`
	WalletID     string       `json:"wallet_id,omitempty"`
	UserID       string       `json:"user_id,omitempty"`
	Type         Type         `json:"type,omitempty"`
	Status       Status       `json:"status,omitempty"`
	CurrencyType string `json:"currency_type,omitempty"`
}

// Wallet keeps representation of the
// user wallet
type Wallet struct {
	DocType  string `json:"doc_type,omitempty"`
	WalletID string `json:"wallet_id,omitempty"`
	UserID   string `json:"user_id,omitempty"`
	Type     Type   `json:"type,omitempty"`
	Status   string `json:"status,omitempty"`
}

type Type string

// Enumerated types values
const (
	Main       Type = "MAIN"
	Additional Type = "ADDITIONAL"
)

// Status encapsulates account/wallet
// statuses values
type Status string

// Enumerate account and wallet statuses
// 1. Blocked - account or wallet being activated
// 3. Active - account/wallet is blocked
const (
	Blocked Status = "BLOCKED"
	Active  Status = "ACTIVE"
)