package mngr

import (
	"fmt"

	"encoding/json"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/hyperledger/fabric/protos/peer"
	log "github.com/sirupsen/logrus"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/actions"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/checker"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/requests"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/tools"
)

var (
	logger = log.WithFields(log.Fields{
		"chaincode": "offerCC",
		"module":    "chaincode",
	})
)

// chaincodeStubInterfaceAdapter interface to encapsulate only required
// methods of shim.ChaincodeStubInterface
type chaincodeStubInterfaceAdapter interface {
	GetState(key string) ([]byte, error)

	PutState(key string, value []byte) error

	GetQueryResult(query string) (shim.StateQueryIteratorInterface, error)

	InvokeChaincode(chaincodeName string, args [][]byte, channel string) peer.Response
}

// offerChaincodeAction represents chaincode actions releted to offer request operations
type offerChaincodeAction func(args []string, adapter chaincodeStubInterfaceAdapter) peer.Response

// offerChaincode offer request chaincode struct
type offerChaincode struct {
	actions map[string]offerChaincodeAction
}

func (c *offerChaincode) Init(stub shim.ChaincodeStubInterface) peer.Response {
	logger.Info("Start initializing chaincode...")
	return shim.Success([]byte("Chaincode initialized successfully"))
}

func (c *offerChaincode) Invoke(stub shim.ChaincodeStubInterface) peer.Response {
	actionName, args := stub.GetFunctionAndParameters()

	if action, ok := c.actions[actionName]; ok {
		return action(args, stub)
	}

	// Cannot process request exiting and logging error
	logger.Errorf("Invalid action name, provided %s", actionName)
	return shim.Error(fmt.Sprintf("Invalid action name, provided %s", actionName))
}

// NewOfferChaincode creating a new chaincode responsible to operate
// on user offer requests
func NewOfferChaincode() shim.Chaincode {
	mngr := &offerManagerImpl{}

	return &offerChaincode{
		actions: map[string]offerChaincodeAction{
			AddOffer: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				logger.Info("addOffer")

				userParameter, systemParameter, err := checker.CheckAccessParameter(stub, args, actions.Actions().OfferCC().AddOffer)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParameter.UserID: %s", systemParameter.UserID)
				logger.Infof("systemParameter.Param: %s", systemParameter.Param)

				logger.Infof("userParameter.UserID: %s", userParameter.UserID)
				logger.Infof("userParameter.Param: %s", userParameter.Param)

				var offerRequest OfferRequest
				err = json.Unmarshal(userParameter.Param, &offerRequest)
				if err != nil {
					errMsg := fmt.Sprintf("Unmarshal userParams to OfferRequest. Error: %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				if offerRequest.OutUserInfo.UserID != userParameter.UserID {
					errMsg := fmt.Sprintf("Can't %s because signer UserID %s not equals with out UserID %s", AddOffer, userParameter.UserID, offerRequest.OutUserInfo.UserID)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				var operationID requests.Id
				if err = json.Unmarshal(systemParameter.Param, &operationID); err != nil {
					errMsg := fmt.Sprintf("Unmarshal systemParams to operationID. Error: %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				offerRequest.OperationID = operationID.Id
				offerRequest.RequestStatus = OfferPending

				accountInfo := AccountInfo{
					UserID:    offerRequest.InUserInfo.UserID,
					WalletID:  offerRequest.InUserInfo.WalletID,
					AccountID: offerRequest.InUserInfo.AccountID,
					Status:    Active,
				}

				if resp := tools.SignedInvokeChaincode(stub, "accountchannel",
					actions.Actions().AccountCC().CheckAndLoadAccount, accountInfo); resp.Status == 200 {

					var account Account
					if err := json.Unmarshal(resp.Payload, &account); err != nil {
						errMsg := fmt.Sprintf("Wasn't able to unmarshal account passed as parameter, due to %s", err)
						logger.Error(errMsg)
						return shim.Error(errMsg)
					}

					if account.UserID == offerRequest.InUserInfo.UserID {
						offerRequest.InUserInfo.WalletID = account.WalletID
						offerRequest.InUserInfo.AccountID = account.AccountID

						logger.Debugf("Adding new offer offerRequest from user %s, operation id %s", offerRequest.OutUserInfo.UserID, offerRequest.OperationID)
						err = mngr.AddOffer(&offerRequest, stub)
						if err != nil {
							logger.Error(err)
							return shim.Error(err.Error())
						}

						offerRequestJson, err := json.Marshal(offerRequest)
						if err != nil {
							errMsg := fmt.Sprintf("Problem Marshal offerRequest. Error: %s", err)
							logger.Error(errMsg)
							return shim.Error(errMsg)
						}

						return shim.Success([]byte(offerRequestJson))
					} else {
						logger.Error("Account validation failed")
						return shim.Error("Account validation failed")
					}

				} else {
					logger.Error(resp.Message)
					return shim.Error(resp.Message)
				}
			},
			ReplayOffer: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				userParameter, systemParameter, err := checker.CheckAccessParameter(stub, args, actions.Actions().OfferCC().ReplayOffer)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParameter.UserID: %s", systemParameter.UserID)
				logger.Infof("systemParameter.Param: %s", systemParameter.Param)

				logger.Infof("userParameter.UserID: %s", userParameter.UserID)
				logger.Infof("userParameter.Param: %s", userParameter.Param)

				var offerRequest OfferRequest
				err = json.Unmarshal(systemParameter.Param, &offerRequest)
				if err != nil {
					errMsg := fmt.Sprintf("Wasn't able to read organization passed as paraemter, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				if offerRequest.OutUserInfo.UserID != userParameter.UserID {
					errMsg := fmt.Sprintf("Can't %s because signer UserID %s not equals with out UserID %s", ReplayOffer, userParameter.UserID, offerRequest.OutUserInfo.UserID)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				logger.Debugf("Adding new offer request from user %s, operation id %s", offerRequest.OutUserInfo.UserID, offerRequest.OperationID)
				err = mngr.ReplayOffer(&offerRequest, stub)
				if err != nil {
					logger.Error(err)
					return shim.Error(err.Error())
				}

				return shim.Success(systemParameter.Param)
			},
			GetReplayHistoryOffer: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().OfferCC().GetReplayHistoryOffer)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParams: %s", systemParams)

				var id Id
				err = json.Unmarshal(systemParams, &id)
				if err != nil {
					errMsg := fmt.Sprintf("Wasn't able to read organization passed as paraemter, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				operationID := id.ID
				logger.Infof("operationID: %s", operationID)

				if len(operationID) == 0 {
					errMsg := "offer operation id was not provided"
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				logger.Debugf("GetReplayHistoryOffer by operation id %s", operationID)
				usersOffers, err := mngr.GetReplayHistoryOffer(operationID, stub)
				if err != nil {
					logger.Error(err)
					return shim.Error(err.Error())
				}

				usersOffersByteArray, err := json.Marshal(*usersOffers)
				if err != nil {
					errMsg := fmt.Sprintf("failed to serialize incoming user offers with status %s, due to %s", usersOffers, err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				return shim.Success(usersOffersByteArray)
			},
			GetOffer: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				logger.Info("getOffer")

				userParams, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().OfferCC().GetOffer)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParams: %s", systemParams)
				logger.Infof("userParams: %s", userParams)

				var id Id
				err = json.Unmarshal(systemParams, &id)
				if err != nil {
					errMsg := fmt.Sprintf("Problem Unmarshal systemParams. Error: %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				operationID := id.ID
				logger.Infof("operationID: %s", operationID)

				if len(operationID) == 0 {
					errMsg := "offer operation id was not provided"
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				bytes, _, err := mngr.GetOffer(operationID, stub)
				if err != nil {
					return shim.Error(err.Error())
				}

				if len(bytes) == 0 {
					msg := fmt.Sprintf("Can't find offer by %s", operationID)
					return shim.Error(msg)
				}

				return shim.Success(bytes)
			},
			ChangeOfferStatusByOperationId: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().OfferCC().ChangeOfferStatusByOperationId)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParams: %s", systemParams)

				var changeOfferStatusByOperationIdArg ChangeOfferStatusByOperationIdArg
				err = json.Unmarshal(systemParams, &changeOfferStatusByOperationIdArg)
				if err != nil {
					errMsg := fmt.Sprintf("Wasn't able to read organization passed as paraemter, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				// expecting zero argument to represent operation id
				operationID := changeOfferStatusByOperationIdArg.OperationID
				if len(operationID) == 0 {
					errMsg := "offer operation id was not provided"
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				offerStatus := OfferRequestStatus(changeOfferStatusByOperationIdArg.OfferRequestStatus)
				if len(offerStatus) == 0 {
					errMsg := fmt.Sprintf("no new offers status provided operation id %s", operationID)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				logger.Debugf("Changing offer request status operation id, source user id %s", operationID)

				bytes, err := mngr.ChangeOfferStatus(operationID, offerStatus, stub)
				if err != nil {
					logger.Error(err)
					return shim.Error(err.Error())
				}

				return shim.Success(bytes)
			},

			GetUserIncomingRequestsByStatus: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().OfferCC().GetUserIncomingRequestsByStatus)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParams: %s", systemParams)

				var getUserIncomingRequestsByStatusArg GetUserIncomingRequestsByStatusArg
				err = json.Unmarshal(systemParams, &getUserIncomingRequestsByStatusArg)
				if err != nil {
					errMsg := fmt.Sprintf("Wasn't able to read organization passed as paraemter, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				userID := getUserIncomingRequestsByStatusArg.UserID
				if len(userID) == 0 {
					errMsg := "empty user id was provided, unable to retrieve incoming requests"
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				offerStatus := OfferRequestStatus(getUserIncomingRequestsByStatusArg.OfferRequestStatus)
				if len(offerStatus) == 0 {
					errMsg := fmt.Sprintf("no offers status provided, cannot get a list of users incoming requests for id %s", userID)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				usersOffers, err := mngr.GetUserIncomingRequestsByStatus(userID, offerStatus, stub)
				if err != nil {
					logger.Error(err)
					return shim.Error(err.Error())
				}

				usersOffersByteArray, err := json.Marshal(*usersOffers)
				if err != nil {
					errMsg := fmt.Sprintf("failed to serialize incoming user offers with status %s for user id %s, due to %s", offerStatus, userID, err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				return shim.Success(usersOffersByteArray)
			},

			GetRequestsByStatus: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().OfferCC().GetRequestsByStatus)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParams: %s", systemParams)

				var getRequestsByStatusArg GetRequestsByStatusArg
				err = json.Unmarshal(systemParams, &getRequestsByStatusArg)
				if err != nil {
					errMsg := fmt.Sprintf("Wasn't able to read organization passed as paraemter, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				offerStatus := OfferRequestStatus(getRequestsByStatusArg.OfferRequestStatus)
				if len(offerStatus) == 0 {
					errMsg := fmt.Sprintf("no offers status provided, cannot get a list of requests with status %s", offerStatus)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				offers, err := mngr.GetRequestsByStatus(offerStatus, stub)
				if err != nil {
					logger.Error(err)
					return shim.Error(err.Error())
				}

				offersJson, err := json.Marshal(*offers)
				if err != nil {
					errMsg := fmt.Sprintf("failed to serialize incoming user offers with status %s, due to %s", offerStatus, err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				return shim.Success([]byte(offersJson))
			},

			GetUserOutgoingRequestsByStatus: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().OfferCC().GetUserOutgoingRequestsByStatus)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParams: %s", systemParams)

				var getUserOutgoingRequestsByStatusArg GetUserOutgoingRequestsByStatusArg
				err = json.Unmarshal(systemParams, &getUserOutgoingRequestsByStatusArg)
				if err != nil {
					errMsg := fmt.Sprintf("Wasn't able to read organization passed as paraemter, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				userID := getUserOutgoingRequestsByStatusArg.UserID
				if len(userID) == 0 {
					errMsg := "empty user id was provided, unable to retrieve outgoing requests"
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				offerStatus := OfferRequestStatus(getUserOutgoingRequestsByStatusArg.OfferRequestStatus)
				if len(offerStatus) == 0 {
					errMsg := fmt.Sprintf("no offers status provided, cannot get a list of users outgoing requests for id %s", userID)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				offers, err := mngr.GetUserOutgoingRequestsByStatus(userID, offerStatus, stub)
				if err != nil {
					logger.Error(err)
					return shim.Error(err.Error())
				}

				offersJson, err := json.Marshal(*offers)
				if err != nil {
					errMsg := fmt.Sprintf("failed to serialize incoming user offers with status %s for user id %s, due to %s", offerStatus, userID, err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				return shim.Success([]byte(offersJson))
			},

			GetHistoryUserOutgoingRequests: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().OfferCC().GetHistoryUserOutgoingRequests)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParams: %s", systemParams)

				var getHistoryUserOutgoingRequestsArg GetHistoryUserOutgoingRequestsArg
				err = json.Unmarshal(systemParams, &getHistoryUserOutgoingRequestsArg)
				if err != nil {
					errMsg := fmt.Sprintf("Wasn't able to read organization passed as paraemter, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				userID := getHistoryUserOutgoingRequestsArg.UserID
				if len(userID) == 0 {
					errMsg := "empty user id was provided, unable to retrieve incoming requests"
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				offers, err := mngr.GetHistoryUserOutgoingRequests(userID, stub)
				if err != nil {
					logger.Error(err)
					return shim.Error(err.Error())
				}

				offersJson, err := json.Marshal(*offers)
				if err != nil {
					errMsg := fmt.Sprintf("failed to serialize incoming user offers for user id %s, due to %s", userID, err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				return shim.Success([]byte(offersJson))
			},

			GetHistoryUserIncomingRequests: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().OfferCC().GetHistoryUserIncomingRequests)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParams: %s", systemParams)

				var getHistoryUserIncomingRequestsArg GetHistoryUserIncomingRequestsArg
				err = json.Unmarshal(systemParams, &getHistoryUserIncomingRequestsArg)
				if err != nil {
					errMsg := fmt.Sprintf("Wasn't able to read organization passed as paraemter, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}
				userID := getHistoryUserIncomingRequestsArg.UserID
				if len(userID) == 0 {
					errMsg := "empty user id was provided, unable to retrieve outgoing requests"
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				offers, err := mngr.GetHistoryUserIncomingRequests(userID, stub)
				if err != nil {
					logger.Error(err)
					return shim.Error(err.Error())
				}

				offersJson, err := json.Marshal(*offers)
				if err != nil {
					errMsg := fmt.Sprintf("failed to serialize incoming user offers for user id %s, due to %s", userID, err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				return shim.Success([]byte(offersJson))
			},

			IsOfferApproved: func(args []string, stub chaincodeStubInterfaceAdapter) peer.Response {
				_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().OfferCC().IsOfferApproved)
				if err != nil {
					return shim.Error(err.Error())
				}

				logger.Infof("systemParams: %s", systemParams)

				var isOfferApprovedArg IsOfferApprovedArg
				err = json.Unmarshal(systemParams, &isOfferApprovedArg)
				if err != nil {
					errMsg := fmt.Sprintf("Wasn't able to read organization passed as paraemter, due to %s", err)
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				operationID := isOfferApprovedArg.OperationID
				if len(operationID) == 0 {
					errMsg := "empty operation id is provided cannot continue"
					logger.Error(errMsg)
					return shim.Error(errMsg)
				}

				if err := mngr.IsOfferApproved(operationID, stub); err != nil {
					logger.Error(err.Error())
					shim.Error(err.Error())
				}

				return shim.Success(nil)
			},
		},
	}
}
