package mngr

import (
	"encoding/json"
	"fmt"
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/nums"
)

// CollectOffers execute quert and collect resulting set into marshable list
// of user's offers
func (o *UsersOffers) CollectOffers(query string, adapter chaincodeStubInterfaceAdapter) error {
	it, err := adapter.GetQueryResult(query)
	if err != nil {
		errMsg := fmt.Sprintf("failed to retreive all outcoming user request for user id %s, due to %s", o.UserID, err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	for it.HasNext() {
		kv, err := it.Next()
		if err != nil {
			errMsg := fmt.Sprintf("failed to get next outcoming user request for user id %s, due to %s", o.UserID, err)
			logger.Error(errMsg)
			return errors.New(errMsg)
		}
		var offer OfferRequest
		err = json.Unmarshal(kv.Value, &offer)
		if err != nil {
			errMsg := fmt.Sprintf("failed to deserialize outcoming user request for user id %s, due to %s", o.UserID, err)
			logger.Error(errMsg)
			return errors.New(errMsg)
		}

		o.Offers = append(o.Offers, &offer)
	}

	return nil
}

// OfferManager manager to operate on
// offers requests comming from the users
type OfferManager interface {
	// AddOffer add new offer into the system
	AddOffer(offer *OfferRequest, adapter chaincodeStubInterfaceAdapter) error

	// ReplayOffer add new offer into the system
	ReplayOffer(offer *OfferRequest, adapter chaincodeStubInterfaceAdapter) error

	// GetReplayHistoryOffer add new offer into the system
	GetReplayHistoryOffer(offer *OfferRequest, adapter chaincodeStubInterfaceAdapter) (*UsersOffers, error)

	// ChangeOfferStatus change user offer status
	ChangeOfferStatus(operationID string, scrUserID string, inUserID string, offerStatus OfferRequestStatus, adapter chaincodeStubInterfaceAdapter) ([]byte, error)

	// GetRequestsByStatus retrieves all offers with given status
	GetRequestsByStatus(offerStatus OfferRequestStatus, adapter chaincodeStubInterfaceAdapter) (*OffersList, error)

	// GetUserIncomingRequestsByStatus get all incoming requests of given user based on offers statuses
	GetUserIncomingRequestsByStatus(userID string, offerStatus OfferRequestStatus, adapter chaincodeStubInterfaceAdapter) (*UsersOffers, error)

	// GetUserOutgoingRequestsByStatus get all outgoing requests of given user based on offers statuses
	GetUserOutgoingRequestsByStatus(userID string, offerStatus OfferRequestStatus, adapter chaincodeStubInterfaceAdapter) (*UsersOffers, error)

	// GetHistoryUserIncomingRequests get all incoming requests of given user based on offers statuses
	GetHistoryUserIncomingRequests(userID string, adapter chaincodeStubInterfaceAdapter) (*UsersOffers, error)

	// GetHistoryUserOutgoingRequests get all outgoing requests of given user based on offers statuses
	GetHistoryUserOutgoingRequests(userID string, adapter chaincodeStubInterfaceAdapter) (*UsersOffers, error)

	// IsOfferApproved checks whenver there is an offered approved for given operation id
	IsOfferApproved(operationID string, adapter chaincodeStubInterfaceAdapter) error

	// UpdateOfferProperty updates property of the offer for given operation id,
	// offer should not be in completed status
	UpdateOfferProperty(operationID string, value string, adapter chaincodeStubInterfaceAdapter) error

	// ChangeOfferStatusByTxID change offer status for given record with given external transaction id
	ChangeOfferStatusByTxID(externalTxID string, offerStatus OfferRequestStatus, adapter chaincodeStubInterfaceAdapter) error
}

// offerManagerImpl implementation of the OfferManager interface
type offerManagerImpl struct {
}

// AddOffer add new offer into the system
func (mngr *offerManagerImpl) AddOffer(offer *OfferRequest, adapter chaincodeStubInterfaceAdapter) error {
	if len(offer.ReplayedID) != 0 {
		errMsg := "ReplayedID need to be empty"
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if len(offer.OperationID) == 0 {
		errMsg := "operation id wasn't not provided"
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if len(offer.OutUserInfo.UserID) == 0 {
		errMsg := fmt.Sprintf("out user id was not provided, cannot place offer with operatio id %s", offer.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if len(offer.InUserInfo.UserID) == 0 {
		errMsg := fmt.Sprintf("in user id was not provided, cannot place offer with operatio id %s", offer.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// check out user account status
	if err := mngr.checkAccountStatus(offer.OutUserInfo, adapter); err != nil {
		return err
	}

	// check in user account status
	if err := mngr.checkAccountStatus(offer.InUserInfo, adapter); err != nil {
		return err
	}

	o, err := adapter.GetState(offer.OperationID)
	if err != nil {
		errMsg := fmt.Sprintf("failed to check whenever offer with operation id %s was already exists, due to %s", offer.OperationID, err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if o != nil {
		errMsg := fmt.Sprintf("offer with operation id %s was already exists", offer.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// check offer status
	if offer.RequestStatus != OfferPending {
		errMsg := fmt.Sprintf("new offer request should be of pending status, operation id %s", offer.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	outLockAmountResult, err := mngr.FindLockAmountByOperationIDAndUserId(offer.OperationID, offer.OutUserInfo.UserID, adapter)
	if err != nil {
		return err
	}
	if !outLockAmountResult.Amount.Number().Equals(offer.OutUserInfo.Amount.Number()) {
		errMsg := fmt.Sprintf("outLockAmountResult. locked amount %s for user %s dose not equalse expected offer amount %s", outLockAmountResult.Amount, offer.OutUserInfo.UserID, offer.OutUserInfo.Amount)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	inLockAmountResult, err := mngr.FindLockAmountByOperationIDAndUserId(offer.OperationID, offer.InUserInfo.UserID, adapter)
	if err != nil {
		return err
	}
	if !inLockAmountResult.Amount.Number().Equals(nums.Zero()) {
		errMsg := fmt.Sprintf("inLockAmountResult. locked amount %s for user %s dose not equalse expected offer amount %s", inLockAmountResult.Amount, offer.InUserInfo.UserID, 0)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	logger.Debugf("adding new offer into ledger, operation id %s, out user id %s, from currency id %s",
		offer.OperationID, offer.OutUserInfo.UserID, offer.OutUserInfo.Amount.CurrencyType)

	offerJSON, err := json.Marshal(offer)
	if err != nil {
		errMsg := fmt.Sprintf("failed to deserialize user offer request, operation id %s, from user id %s, due to %s",
			offer.OperationID, offer.OutUserInfo.UserID, err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Adding new offer into ledger state
	return adapter.PutState(offer.OperationID, offerJSON)
}

// ReplayOffer add new offer into the system
func (mngr *offerManagerImpl) ReplayOffer(offer *OfferRequest, adapter chaincodeStubInterfaceAdapter) error {
	if len(offer.ReplayedID) == 0 {
		errMsg := "ReplayedID wasn't not provided"
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if len(offer.OperationID) == 0 {
		errMsg := "OperationID wasn't not provided"
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if len(offer.OutUserInfo.UserID) == 0 {
		errMsg := fmt.Sprintf("out user id was not provided, cannot place offer with operatio id %s", offer.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if len(offer.InUserInfo.UserID) == 0 {
		errMsg := fmt.Sprintf("in user id was not provided, cannot place offer with operatio id %s", offer.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// check out user account status
	if err := mngr.checkAccountStatus(offer.OutUserInfo, adapter); err != nil {
		return err
	}

	// check in user account status
	if err := mngr.checkAccountStatus(offer.InUserInfo, adapter); err != nil {
		return err
	}

	o, err := adapter.GetState(offer.OperationID)
	if err != nil {
		errMsg := fmt.Sprintf("failed to check whenever offer with operation id %s was already exists, due to %s", offer.OperationID, err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if o != nil {
		errMsg := fmt.Sprintf("offer with operation id %s was already exists", offer.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// check offer status
	if offer.RequestStatus != OfferPending {
		errMsg := fmt.Sprintf("new offer request should be of pending status, operation id %s", offer.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	outLockAmountResult, err := mngr.FindLockAmountByOperationIDAndUserId(offer.OperationID, offer.OutUserInfo.UserID, adapter)
	if err != nil {
		return err
	}
	if outLockAmountResult.Amount != offer.OutUserInfo.Amount {
		errMsg := fmt.Sprintf("outLockAmountResult. locked amount %s for user %s dose not equalse expected offer amount %s", outLockAmountResult.Amount, offer.OutUserInfo.UserID, offer.OutUserInfo.Amount)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	inLockAmountResult, err := mngr.FindLockAmountByOperationIDAndUserId(offer.OperationID, offer.InUserInfo.UserID, adapter)
	if err != nil {
		return err
	}
	if !inLockAmountResult.Amount.Number().Equals(nums.Zero()) {
		errMsg := fmt.Sprintf("inLockAmountResult. locked amount %s for user %s dose not equalse expected offer amount %s", inLockAmountResult.Amount, offer.InUserInfo.UserID, 0)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	logger.Debugf("adding new offer into ledger, operation id %s, out user id %s, from currency id %s",
		offer.OperationID, offer.OutUserInfo.UserID, offer.OutUserInfo.Amount.CurrencyType)

	offerJSON, err := json.Marshal(offer)
	if err != nil {
		errMsg := fmt.Sprintf("failed to deserialize user offer request, operation id %s, from user id %s, due to %s",
			offer.OperationID, offer.OutUserInfo.UserID, err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Adding new offer into ledger state
	return adapter.PutState(offer.OperationID, offerJSON)
}

// ReplayOffer add new offer into the system
func (mngr *offerManagerImpl) GetReplayHistoryOffer(operationID string, stub chaincodeStubInterfaceAdapter) (*UsersOffers, error) {
	result := &UsersOffers{
		UserID: "unknown",
		Offers: make([]*OfferRequest, 0),
	}
	if len(operationID) == 0 {
		errMsg := "operation id wasn't not provided"
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	_, offerRequest, err := mngr.GetOffer(operationID, stub)
	if err != nil {
		logger.Errorf("error: " + err.Error())
		return nil, err
	}

	for i := 0; ; i++ {
		result.Offers = append(result.Offers, offerRequest)
		if len(offerRequest.ReplayedID) != 0 {
			_, offerRequest, err = mngr.GetOffer(operationID, stub)
			if err != nil {
				break
			}
		}
	}

	// Adding new offerByteArray into ledger state
	return result, nil
}

// ChangeOfferStatus change user offer status
func (mngr *offerManagerImpl) ChangeOfferStatus(operationID string, offerStatus OfferRequestStatus, adapter chaincodeStubInterfaceAdapter) ([]byte, error) {
	offerJSON, err := adapter.GetState(operationID)
	if err != nil {
		errMsg := fmt.Sprintf("failed to obtain offer with operation id %s was already exists, due to %s", operationID, err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	if len(offerJSON) == 0 {
		errMsg := fmt.Sprintf("cannot find offer by operationId %s", operationID)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	var offer OfferRequest
	json.Unmarshal(offerJSON, &offer)

	// check out user account status
	if err := mngr.checkAccountStatus(offer.OutUserInfo, adapter); err != nil {
		return nil, err
	}

	// check in user account status
	if err := mngr.checkAccountStatus(offer.InUserInfo, adapter); err != nil {
		return nil, err
	}

	if offerStatus == OfferAccepted {
		outLockAmountResult, err := mngr.FindLockAmountByOperationIDAndUserId(offer.OperationID, offer.OutUserInfo.UserID, adapter)
		if err != nil {
			return nil, err
		}
		if outLockAmountResult.Amount != offer.OutUserInfo.Amount {
			errMsg := fmt.Sprintf("loutLockAmountResult. locked amount %s for user %s dose not equalse expected offer amount %s", outLockAmountResult.Amount, offer.OutUserInfo.UserID, offer.OutUserInfo.Amount)
			logger.Error(errMsg)
			return nil, errors.New(errMsg)
		}

		inLockAmountResult, err := mngr.FindLockAmountByOperationIDAndUserId(offer.OperationID, offer.InUserInfo.UserID, adapter)
		if err != nil {
			return nil, err
		}
		if inLockAmountResult.Amount != offer.InUserInfo.Amount {
			errMsg := fmt.Sprintf("inLockAmountResult. locked amount %s for user %s dose not equalse expected offer amount %s", inLockAmountResult.Amount, offer.InUserInfo.UserID, offer.InUserInfo.Amount)
			logger.Error(errMsg)
			return nil, errors.New(errMsg)
		}
	} else if offerStatus == OfferPending {
		errMsg := fmt.Sprintf("can't change offer status to pending")
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	logger.Debugf("updating offer status, operation id %s, out user id %s, from currency id %s, new status",
		offer.OperationID, offer.OutUserInfo.UserID, offer.OutUserInfo.Amount.CurrencyType, offerStatus)

	offer.RequestStatus = offerStatus
	offerJSON, err = json.Marshal(offer)
	if err != nil {
		errMsg := fmt.Sprintf("failed to deserialize user offer request, operation id %s, from user id %s, due to %s",
			offer.OperationID, offer.OutUserInfo.UserID, err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	// Update status
	err = adapter.PutState(offer.OperationID, offerJSON)
	return offerJSON, err
}

func (mngr *offerManagerImpl) GetOffer(operationID string, adapter chaincodeStubInterfaceAdapter) ([]byte, *OfferRequest, error) {
	offerJSON, err := adapter.GetState(operationID)
	if err != nil {
		errMsg := fmt.Sprintf("failed to obtain offer with operation id %s was already exists, due to %s", operationID, err)
		logger.Error(errMsg)
		return nil, nil, errors.New(errMsg)
	}

	if len(offerJSON) == 0 {
		errMsg := fmt.Sprintf("cannot find offer by operationId %s", operationID)
		logger.Error(errMsg)
		return nil, nil, errors.New(errMsg)
	}

	var offer OfferRequest
	err = json.Unmarshal(offerJSON, &offer)
	if err != nil {
		errMsg := fmt.Sprintf("Error json unmarshal. %s", err.Error())
		logger.Error(errMsg)
		return nil, nil, errors.New(errMsg)
	}

	return offerJSON, &offer, err
}

// GetUserIncomingRequestsByStatus get all incoming requests of given user based on offers statuses
func (mngr *offerManagerImpl) GetUserIncomingRequestsByStatus(userID string, offerStatus OfferRequestStatus, adapter chaincodeStubInterfaceAdapter) (*UsersOffers, error) {
	result := &UsersOffers{
		UserID: userID,
		Offers: make([]*OfferRequest, 0),
	}

	query, err := getSelector(InOfferRequest{
		InUserInfo: OfferRequestInfo{
			UserID: userID,
		},
		RequestStatus: offerStatus,
	})
	if err != nil {
		errMsg := fmt.Sprintf("failed to create query. Error: %s", err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	logger.Info("query: ", query)

	if err := result.CollectOffers(query, adapter); err != nil {
		return nil, err
	}

	return result, nil
}

// GetUserOutgoingRequestsByStatus get all outgoing requests of given user based on offers statuses
func (mngr *offerManagerImpl) GetUserOutgoingRequestsByStatus(userID string, offerStatus OfferRequestStatus, adapter chaincodeStubInterfaceAdapter) (*UsersOffers, error) {
	result := &UsersOffers{
		UserID: userID,
		Offers: make([]*OfferRequest, 0),
	}

	query, err := getSelector(OutOfferRequest{
		OutUserInfo: OfferRequestInfo{
			UserID: userID,
		},
		RequestStatus: offerStatus,
	})
	if err != nil {
		errMsg := fmt.Sprintf("failed to create query. Error: %s", err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	logger.Info("query: ", query)

	if err := result.CollectOffers(query, adapter); err != nil {
		return nil, err
	}

	return result, nil
}

// GetUserIncomingRequestsByStatus get all incoming requests of given user based on offers statuses
func (mngr *offerManagerImpl) GetHistoryUserIncomingRequests(userID string, adapter chaincodeStubInterfaceAdapter) (*UsersOffers, error) {
	result := &UsersOffers{
		UserID: userID,
		Offers: make([]*OfferRequest, 0),
	}

	query, err := getSelector(InOfferRequest{
		InUserInfo: OfferRequestInfo{
			UserID: userID,
		},
	})
	if err != nil {
		errMsg := fmt.Sprintf("failed to create query. Error: %s", err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	logger.Info("query: ", query)

	if err := result.CollectOffers(query, adapter); err != nil {
		return nil, err
	}

	return result, nil
}

// GetUserOutgoingRequestsByStatus get all outgoing requests of given user based on offers statuses
func (mngr *offerManagerImpl) GetHistoryUserOutgoingRequests(userID string, adapter chaincodeStubInterfaceAdapter) (*UsersOffers, error) {
	result := &UsersOffers{
		UserID: userID,
		Offers: make([]*OfferRequest, 0),
	}

	query, err := getSelector(OutOfferRequest{
		OutUserInfo: OfferRequestInfo{
			UserID: userID,
		},
	})
	if err != nil {
		errMsg := fmt.Sprintf("failed to create query. Error: %s", err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	logger.Info("query: ", query)

	if err := result.CollectOffers(query, adapter); err != nil {
		return nil, err
	}

	return result, nil
}

// IsOfferApproved checks whenver there is an offered approved for given operation id
func (mngr *offerManagerImpl) IsOfferApproved(operationID string, adapter chaincodeStubInterfaceAdapter) error {

	query, err := getSelector(OfferRequest{
		OperationID:   operationID,
		RequestStatus: OfferAccepted,
	})

	if err != nil {
		errMsg := fmt.Sprintf("failed to create query selector, due to %s", err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	logger.Debugf("Checking whenever there is an approved offer, for operation id %s", operationID)
	it, err := adapter.GetQueryResult(query)
	if err != nil {
		errMsg := fmt.Sprintf("cannot read from state db due to %s", err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if !it.HasNext() {
		errMsg := fmt.Sprintf("no apporved offer for operation id %s found", operationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	return nil
}

// GetRequestsByStatus returns list of the
func (mngr *offerManagerImpl) GetRequestsByStatus(offerStatus OfferRequestStatus, adapter chaincodeStubInterfaceAdapter) (*OffersList, error) {
	var result OffersList

	query, err := getSelector(OutOfferRequest{
		RequestStatus: offerStatus,
	})
	if err != nil {
		errMsg := fmt.Sprintf("failed to create query. Error: %s", err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	logger.Debug("query: ", query)

	it, err := adapter.GetQueryResult(query)
	if err != nil {
		errMsg := fmt.Sprintf("failed to retreive all request with offer status %s, due to %s", offerStatus, err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	for it.HasNext() {
		kv, err := it.Next()
		if err != nil {
			errMsg := fmt.Sprintf("failed to get next outcoming user request with offer status %s, due to %s", offerStatus, err)
			logger.Error(errMsg)
			return nil, errors.New(errMsg)
		}
		var offer OfferRequest
		err = json.Unmarshal(kv.Value, &offer)
		if err != nil {
			errMsg := fmt.Sprintf("failed to deserialize request with offer status id %s, due to %s", offerStatus, err)
			logger.Error(errMsg)
			return nil, errors.New(errMsg)
		}

		result.Offers = append(result.Offers, &offer)
	}

	return &result, nil

}

func getSelector(data interface{}) (string, error) {
	query, err := json.Marshal(struct {
		Selector interface{} `json:"selector"`
	}{
		Selector: data,
	})

	return string(query), err
}
