package mngr

import (
	"encoding/json"
	"fmt"
	"github.com/hyperledger/fabric/common/util"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/hyperledger/fabric/protos/peer"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/mock"
	"testing"
)

// chaincodeStubInterfaceAdapterMock mock implementation of the
// chaincodeStubInterfaceAdapter interface to be used in tests
// to simulate different logic
type chaincodeStubInterfaceAdapterMock struct {
	mock.Mock
}

func (stub *chaincodeStubInterfaceAdapterMock) GetState(key string) ([]byte, error) {
	args := stub.Called(key)
	if args.Get(0) == nil && args.Error(1) == nil {
		return nil, nil
	}
	return args.Get(0).([]byte), args.Error(1)
}

func (stub *chaincodeStubInterfaceAdapterMock) PutState(key string, value []byte) error {
	args := stub.Called(key, value)
	return args.Error(0)
}

func (stub *chaincodeStubInterfaceAdapterMock) GetQueryResult(query string) (shim.StateQueryIteratorInterface, error) {
	args := stub.Called(query)
	return args.Get(0).(shim.StateQueryIteratorInterface), args.Error(1)
}

func (stub *chaincodeStubInterfaceAdapterMock) InvokeChaincode(chaincodeName string, args [][]byte, channel string) peer.Response {
	res := stub.Called(chaincodeName, args, channel)
	return res.Get(0).(peer.Response)
}

func TestOfferManagerImpl_AddValidOffer(t *testing.T) {
	// Initialize offer manager
	mngr := offerManagerImpl{}

	testOperationID := "offerID1"

	offer := &OfferRequest{
		OperationID: testOperationID,
		OutUserInfo: OfferRequestInfo{
			UserID:       "outUserID",
			WalletID:     "outWalletID",
			AccountID:    "outAccountID",
			CurrencyType: "BTC",
			Amount:       100.0,
		},
		InUserInfo: OfferRequestInfo{
			UserID:       "inUserID",
			WalletID:     "inWalletID",
			AccountID:    "inAccountID",
			CurrencyType: "LTC",
			Amount:       100.0, // TODO: Do we really need amount for in user having rate value in offer request?
		},
		RequestStatus: OfferPending,
	}

	adapter := &chaincodeStubInterfaceAdapterMock{}

	adapter.On("GetState", testOperationID).Return(nil, nil)

	userJSON, err := json.Marshal(offer.OutUserInfo.toUserAccountInfo())
	assert.NoError(t, err)

	args := util.ToChaincodeArgs(VerifyAccountFunctionName, string(userJSON))
	adapter.On("InvokeChaincode", VerifyAccountChaincodeName, args, VerifyAccountChannelName).Return(shim.Success(nil))

	userJSON, err = json.Marshal(offer.InUserInfo.toUserAccountInfo())
	assert.NoError(t, err)

	args = util.ToChaincodeArgs(VerifyAccountFunctionName, string(userJSON))
	adapter.On("InvokeChaincode", VerifyAccountChaincodeName, args, VerifyAccountChannelName).Return(shim.Success(nil))

	offerJson, err := json.Marshal(offer)
	assert.NoError(t, err)

	adapter.On("PutState", testOperationID, offerJson).
		Run(func(args mock.Arguments) {
			// Assert that there was not change while offer data was processed
			// the offer itself
			jo := args.Get(1)
			assert.NotNil(t, jo)
			var request OfferRequest
			err := json.Unmarshal(jo.([]byte), &request)
			assert.NoError(t, err)
			assert.Equal(t, offer, request)
		}).
		Return(nil)

	err = mngr.AddOffer(offer, adapter)
	assert.NoError(t, err)
}

func TestOfferManagerImpl_Unmarshal(t *testing.T) {
	userJSON := "{\"operation_id\":\"OPERATION_ID\",\"out_info\":{\"user_id\":\"USER_ID\",\"wallet_id\":\"WALLET_ID\",\"account_id\":\"ACCOUNT_ID\",\"currency_type\":\"BTC\",\"amount\":1.0},\"in_info\":{\"user_id\":\"USER_ID\",\"wallet_id\":\"WALLET_ID\",\"account_id\":\"ACCOUNT_ID\",\"currency_type\":\"ETH\",\"amount\":2.0},\"request_status\":\"OfferPending\"}"

	var request *OfferRequest = &OfferRequest{}
	err := json.Unmarshal([]byte(userJSON), request)
	if err != nil {
		errMsg := fmt.Sprintf("failed to deserialize user offer request, due to error %s", err)
		logger.Error(errMsg)
	}

	assert.Equal(t, "OPERATION_ID", request.OperationID)
}
