package mngr

import (
	"encoding/json"
	"fmt"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/actions"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/tools"
)

// toUserAccountInfo converts OfferRequestInfo into UserAccountInfo
// representation
func (o OfferRequestInfo) toUserAccountInfo() UserAccountInfo {
	return UserAccountInfo{
		UserID: o.UserID,
		AccountInfo: AccountInfo{
			WalletID:  o.WalletID,
			AccountID: o.AccountID,
		},
	}
}

// checkAccountStatus call account chaincode to verify account status
func (mngr *offerManagerImpl) checkAccountStatus(info OfferRequestInfo, adapter chaincodeStubInterfaceAdapter) error {
	// Check out account status
	resp := tools.SignedInvokeChaincode(adapter, VerifyAccountChannelName, actions.Actions().AccountCC().VerifyUserAccount, info.toUserAccountInfo())

	if resp.Status != int32(shim.OK) {
		errMsg := fmt.Sprintf("verification of user id %s, failed due to %s", info.UserID, resp.Message)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	return nil
}

// получение заблокированных средств по operationID
func (mngr *offerManagerImpl) FindLockAmountByOperationIDAndUserId(operationID string, userID string, adapter chaincodeStubInterfaceAdapter) (*FindLockAmountResult, error) {
	resp := tools.SignedInvokeChaincode(adapter, FindLockAmountByOperationIDAndUserIdChannelName,
		actions.Actions().TransactionCC().FindLockAmountByOperationIDAndUserId, FindLockAmountByOperationIDAndUserIdArg{
			OperationID: operationID,
			UserID:      userID,
		})

	if resp.Status != int32(shim.OK) {
		errMsg := fmt.Sprintf("Problem to find lock transaction by operationID %s and userID %s. Message: %s", operationID, userID, resp.Message)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	var findLockAmountResult FindLockAmountResult
	err := json.Unmarshal(resp.Payload, &findLockAmountResult)
	if err != nil {
		errMsg := fmt.Sprintf("Failed to deserialize response from 'findLockAmountByOperationIDAndUserId' function, due to %s", err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	return &findLockAmountResult, nil
}
