package chaincode

import (
	pb "github.com/hyperledger/fabric/protos/peer"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/sirupsen/logrus"
	"fmt"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/checker"
	"gitlab.smartblocklab.com/smartcontracts/sc-security/requests"
)

var logger = logrus.WithFields(logrus.Fields{
	"module":  "common",
	"package": "chaincode",
})

// ChaincodeStubInterfaceAdapter stub interface to encapsulate only
// relevant parts of the shim.ChaincodeStubInterface API allowing to mock
// this functionality
type StubInterfaceAdapter interface {
	// GetState returns the value of the specified `key` from the
	// ledger. Note that GetState doesn't read data from the writeset, which
	// has not been committed to the ledger. In other words, GetState doesn't
	// consider data modified by PutState that has not been committed.
	// If the key does not exist in the state database, (nil, nil) is returned.
	GetState(key string) ([]byte, error)

	// PutState puts the specified `key` and `value` into the transaction's
	// writeset as a data-write proposal. PutState doesn't effect the ledger
	// until the transaction is validated and successfully committed.
	// Simple keys must not be an empty string and must not start with null
	// character (0x00), in order to avoid range query collisions with
	// composite keys, which internally get prefixed with 0x00 as composite
	// key namespace.
	PutState(key string, value []byte) error

	// GetQueryResult enables to run ricj queries over choudb state db
	// needed to being able effectively extract account information for
	// user, based only on partial parameters available
	GetQueryResult(query string) (shim.StateQueryIteratorInterface, error)

	// InvokeChaincode locally calls the specified chaincode `Invoke` using the
	// same transaction context; that is, chaincode calling chaincode doesn't
	// create a new transaction message.
	// If the called chaincode is on the same channel, it simply adds the called
	// chaincode read set and write set to the calling transaction.
	// If the called chaincode is on a different channel,
	// only the Response is returned to the calling chaincode; any PutState calls
	// from the called chaincode will not have any effect on the ledger; that is,
	// the called chaincode on a different channel will not have its read set
	// and write set applied to the transaction. Only the calling chaincode's
	// read set and write set will be applied to the transaction. Effectively
	// the called chaincode on a different channel is a `Query`, which does not
	// participate in state validation checks in subsequent commit phase.
	// If `channel` is empty, the caller's channel is assumed.
	InvokeChaincode(chaincodeName string, args [][]byte, channel string) pb.Response

	GetTxID() string
}

var CommonChaincodeActionVar = map[string]CommonChaincodeAction{}

var ChaincodeVar = CommonChaincode{
	Actions: CommonChaincodeActionVar,
}

type CommonChaincodeAction func(userParameter *requests.Parameter, systemParameter *requests.Parameter, stub StubInterfaceAdapter, timeGetter TimeGetter) pb.Response

// UserChaincode chaincode capable to operate and manager
// users records, namely add, get and update user records
type CommonChaincode struct {
	Actions map[string]CommonChaincodeAction
}

// Init initialize chaincode with proper parameters
func (t CommonChaincode) Init(stub shim.ChaincodeStubInterface) pb.Response {
	logger.Info("Start initializing ..")
	// has to be done either in this channel or in channel which operates with
	// organization information
	return shim.Success([]byte("Chaincode initialized successfully"))

}

var ChaincodeName string

// Invoke dispatch to the specific chaincode function based on the input parameters
func (t CommonChaincode) Invoke(stub shim.ChaincodeStubInterface) pb.Response {
	functionName, args := stub.GetFunctionAndParameters()
	logger.Infof("ChaincodeName: %s", ChaincodeName)
	logger.Infof("FunctionName: %s", functionName)

	userParameter, systemParameter, err := checker.CheckAccessParameterByArgsAndFunctionName(args, ChaincodeName, functionName, stub)
	if err != nil {
		logger.Error(err.Error())
		return shim.Error(err.Error())
	}

	if userParameter != nil {
		logger.Infof("userParameter.UserID: %s", userParameter.UserID)
		logger.Infof("userParameter.Param: %s", userParameter.Param)
	}

	if systemParameter != nil {
		logger.Infof("systemParameter.UserID: %s", systemParameter.UserID)
		logger.Infof("systemParameter.Param: %s", systemParameter.Param)
	}

	timeFuncFactory := func(stub shim.ChaincodeStubInterface) func() (int64, error) {
		return func() (int64, error) {
			tt, err := stub.GetTxTimestamp()
			if err != nil {
				return 0, err
			}
			return tt.Seconds, nil
		}
	}

	// taking care to dispatch to the correct function based on the input parameter
	if action, ok := t.Actions[functionName]; ok {
		return action(userParameter, systemParameter, stub, timeFuncFactory(stub))
	}

	// No action define with provided function name, reporting error
	invalidFuncName := fmt.Sprintf("Invalid invoke function name, function name provided %s ", functionName)
	logger.Errorf(invalidFuncName)
	return shim.Error(invalidFuncName)
}

type TimeGetter func() (int64, error)


