package nums

import (
	"errors"
	"fmt"
	"github.com/shopspring/decimal"
	"math"
	"strconv"
)

func NewAsset(symbol string, precision uint, maxValue string, minValue string) (*Asset, error) {
	mxv, err := decimal.NewFromString(maxValue)
	if err != nil {
		return nil, err
	}

	mnv, err := decimal.NewFromString(minValue)
	if err != nil {
		return nil, err
	}

	base := precision

	return &Asset{symbol, base, mxv, mnv}, nil
}

func (obj *Asset) String() string {
	return obj.CurrencyType
}

func (obj *Asset) Equals(o2 *Asset) bool {
	return obj.CurrencyType == o2.CurrencyType && obj.Scale == o2.Scale
}

func (obj *CurrencyNumber) Plain() PlainCurrencyNumber {
	return PlainCurrencyNumber{
		Amount:       obj.String(),
		CurrencyType: obj.CurrencyType(),
	}
}

func (obj *PlainCurrencyNumber) Number() *CurrencyNumber {
	return NumFromString(obj.Amount, obj.CurrencyType)
}

func (obj *PlainCurrencyNumber) String() string {
	return NumFromString(obj.Amount, obj.CurrencyType).String()
}

func (obj *CurrencyNumber) CurrencyType() string {
	return obj.Asset.String()
}

func (obj *CurrencyNumber) Copy() *CurrencyNumber {
	return NumFromString(obj.String(), obj.Asset.CurrencyType)
}

func (obj *CurrencyNumber) AssetEquals(b *CurrencyNumber) bool {
	if obj.Asset.Equals(b.Asset) || obj.Asset.Equals(AssetFrom(ZeroAsset)) || b.Asset.Equals(AssetFrom(ZeroAsset)) {
		return true
	} else {
		logger.Errorf(fmt.Sprintf("Assets are different: self=%s, target=%s", obj.CurrencyType(), b.CurrencyType()))
		return false
	}
}

func (obj *CurrencyNumber) UpdateAsset(b *CurrencyNumber) *CurrencyNumber {
	obj.Asset = b.Asset
	return obj
}

func (obj *CurrencyNumber) Validate() error {
	return validate(obj).Error
}

func (obj *CurrencyNumber) String() string {
	return obj.Num.RoundBank(int32(obj.Asset.Scale)).StringFixed(int32(obj.Asset.Scale))
}

func (obj *CurrencyNumber) HasErrors() bool {
	return obj.Error != nil
}

func (obj *CurrencyNumber) ErrorMsg() string {
	return obj.Error.Error()
}

func FinRound(n float64, scale uint) float64 {
	offset := math.Pow(10, float64(scale))
	n *= offset

	x, d := math.Modf(n)

	if d == 0.5 {
		if (int(x)+1)%2 == 0 {
			n = math.Ceil(n)
		} else {
			n = math.Floor(n)
		}
	} else if d > 0.5 {
		n = math.Ceil(n)
	} else {
		n = math.Floor(n)
	}

	return n / offset
}

func MinNumber(a *CurrencyNumber, b *CurrencyNumber) *CurrencyNumber {
	if a.Num.Cmp(b.Num) < 0 {
		return a
	} else {
		return b
	}
}

func MinUInt(a uint64, b uint64) uint64 {
	return uint64(math.Min(float64(a), float64(b)))
}

func correctAsset(a *CurrencyNumber, b *CurrencyNumber) *Asset {
	if !a.Asset.Equals(AssetFrom(ZeroAsset)) {
		return a.Asset
	}

	if !b.Asset.Equals(AssetFrom(ZeroAsset)) {
		return b.Asset
	}

	return AssetFrom(ErrorAssset)
}

func FormatFloat(f float64, scale uint) string {
	return strconv.FormatFloat(f, 'f', int(scale), 64)
}

func validate(res *CurrencyNumber) *CurrencyNumber {
	flag, err := decimal.NewFromString("-1")

	if res.Num.LessThan(decimal.Zero) ||
		((res.Num.GreaterThan(res.Asset.MaxValue) && !res.Asset.MaxValue.Equals(flag)) ||
			(res.Num.LessThan(res.Asset.MinValue) && !res.Asset.MinValue.Equals(flag))) &&
			!res.Num.Equals(decimal.Zero) {

		err = errors.New(fmt.Sprintf("Value is out of bounds [%s...%s]:  %s",
			res.Asset.MinValue.RoundBank(int32(res.Asset.Scale)).StringFixed(int32(res.Asset.Scale)),
			res.Asset.MaxValue.RoundBank(int32(res.Asset.Scale)).StringFixed(int32(res.Asset.Scale)),
			res.Num.StringFixed(int32(res.Asset.Scale))))

		res.Num = decimal.Zero
		res.Error = err
	}

	return res
}

func NumFromString(n string, currencyType string) *CurrencyNumber {
	asset := AssetFrom(currencyType)

	if asset.Equals(AssetFrom(ErrorAssset)) {
		logger.Errorf(fmt.Sprintf("failed to create number %s with bad currencyType %s", n, currencyType))
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: asset,
			Error: errors.New("failed to create ZERO-number with non-zero value"),
		}
	}

	if asset.Equals(AssetFrom(ZeroAsset)) && n != "0" {
		logger.Errorf(fmt.Sprintf("Failed to create ZERO-number with non-zero value: n=%s, asset=%v", n, asset))
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: asset,
			Error: errors.New("failed to create ZERO-number with non-zero value"),
		}
	}

	if num, err := decimal.NewFromString(n); err != nil {
		logger.Errorf(fmt.Sprintf("Failed to create CurrencyNumber, bad string: %s, %v", n, asset))

		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: asset,
			Error: err,
		}
	} else if num.Sign() == -1 {
		logger.Errorf(fmt.Sprintf("Negative numbers are not allowed: n=%s, asset=%v", n, asset))
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: asset,
			Error: errors.New("negative numbers are not allowed"),
		}
	} else {
		return &CurrencyNumber{
			Num:   num,
			Asset: asset,
		}
	}
}

func NumFromFloat(n float64, currencyType string) *CurrencyNumber {
	asset := AssetFrom(currencyType)

	if asset.Equals(AssetFrom(ErrorAssset)) {
		logger.Errorf(fmt.Sprintf("failed to create number %f with bad currencyType %s", n, currencyType))
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: asset,
			Error: errors.New("failed to create ZERO-number with non-zero value"),
		}
	}

	if asset.Equals(AssetFrom(ZeroAsset)) && n != 0 {
		logger.Errorf(fmt.Sprintf("Failed to create ZERO-number with non-zero value: n=%d, asset=%v", n, asset))
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: asset,
			Error: errors.New("failed to create ZERO-number with non-zero value"),
		}
	}

	if n < 0 {
		logger.Errorf(fmt.Sprintf("Failed to create number, negative numbers are not allowed: n=%d, asset=%v", n, asset))
		return &CurrencyNumber{
			Num:   decimal.Zero,
			Asset: asset,
			Error: errors.New("negative numbers are not allowed"),
		}
	}

	return &CurrencyNumber{
		Num:   decimal.NewFromFloat(n),
		Asset: asset,
	}
}

func NumFromInt(n uint64, currencyType string) *CurrencyNumber {
	return NumFromFloat(float64(n), currencyType)
}

func One(n *CurrencyNumber) *CurrencyNumber {
	return NumFromFloat(1, n.Asset.CurrencyType)
}

func Zero() *CurrencyNumber {
	return NumFromFloat(0, ZeroAsset)
}
