package txnmngr

import (
	"encoding/json"
	"errors"
	"fmt"
	"strconv"

	"encoding/binary"
	"github.com/hyperledger/fabric/common/util"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/hyperledger/fabric/protos/peer"
	log "github.com/sirupsen/logrus"
	"gitlab.smartblocklab.com/smartcontracts/common/commissionCC"
)

var (
	logger = log.WithFields(log.Fields{
		"chaincode": "transactionCC",
		"module":    "transactionManager",
	})

	// TODO: retrieve system user id
	systemUser = UserAccountInfo{
		// Keep default values for time being, has to be replaced with
		// call to system chaincode to obtain system user and account information
		UserID: "systemUserID",
		AccountInfo: AccountInfo{
			AccountID: "systemAccountID",
			WalletID:  "systemWalletID",
		}}
)

// chaincodeStubInterfaceAdapter encapsulates ChaincodeStubInterface API
// required for this chaincode
type chaincodeStubInterfaceAdapter interface {
	GetState(key string) ([]byte, error)

	PutState(key string, value []byte) error

	GetTxID() string

	GetQueryResult(query string) (shim.StateQueryIteratorInterface, error)

	InvokeChaincode(chaincodeName string, args [][]byte, channel string) peer.Response
}

// TransactionManager interface to encapture all transaction
// related business processes
type TransactionManager interface {
	// LockFundsOnAccount lock currency funds on specific account
	LockFundsOnAccount(txnArgs *TxnFundsLockInput, adapter *managerAux) error

	// ReleaseFunds lock currency funds on specific account
	ReleaseFunds(txnArgs *TxnFundsLockInput, adapter *managerAux) error

	// AccountCurrencyBalance compute user balance for given currency
	AccountCurrencyBalance(accountID string, currencyType string, adapter *managerAux) (float32, error)

	// TransferFunds finalize transferring funds from locked account
	TransferFunds(txnArgs *TxnFundsLockInput, adapter *managerAux) error

	LoadCryptocurrency(cryptocurrencyTypeInput CryptocurrencyTypeInput, adapter *managerAux) (*ImportCryptocurrencyResult, error)
}

// AccountVerifier verifies account validity
type AccountVerifier func(userAccount *UserAccountInfo, adapter chaincodeStubInterfaceAdapter) (*UserAccountInfo, error)

type CryptocurrencyLoader func(cryptocurrencyTypeInput CryptocurrencyTypeInput, adapter chaincodeStubInterfaceAdapter) (*CryptocurrencyStoreOutput, error)

// transactionManagerImpl implementation of the TransactionManager
// interface which provides functionality claimed
type transactionManagerImpl struct {
	// Account verifier to check account information
	accountVerifier AccountVerifier

	cryptocurrencyLoader CryptocurrencyLoader
}

// NewTransactionManager creates TransactionManager instance wuth account verifier and loader
func NewTransactionManager(verifier AccountVerifier, cryptocurrencyLoader CryptocurrencyLoader) TransactionManager {
	return &transactionManagerImpl{
		accountVerifier:      verifier,
		cryptocurrencyLoader: cryptocurrencyLoader,
	}
}

// LockFundsOnAccount lock funds for transfer from given accountID
func (txnMngr *transactionManagerImpl) LockFundsOnAccount(txnArgs *TxnFundsLockInput, adapter *managerAux) error {
	// ASSUMPTION: All user ids are present, otherwise chaincode invocation will results with error
	if len(txnArgs.FromUser.UserID) == 0 {
		errMsg := fmt.Sprintf("missing from user id, operation id %s", txnArgs.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if len(txnArgs.ToUser.UserID) == 0 {
		errMsg := fmt.Sprintf("missing to user id, operation id %s", txnArgs.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Verify to user information
	fromUser, err := txnMngr.checkAndLoadUser(&txnArgs.FromUser, adapter)
	if err != nil {
		logger.Error(err)
		return err
	}

	// Verify from user information
	// TODO: Why do we need to know the to user if only locking funds for from user w/o storing any information about
	// for which user we are doing it?
	_, err = txnMngr.checkAndLoadUser(&txnArgs.ToUser, adapter)
	if err != nil {
		logger.Error(err)
		return err
	}

	if txnArgs.Amount <= 0 {
		errMsg := fmt.Sprintf("funds amount to lock should be positive number, got %f", txnArgs.Amount)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Get currency balance for user transferring funds from
	logger.Debugf("Computing balance for account id %s, currency id %s", fromUser.AccountID, txnArgs.CurrencyType)
	currencyBalance, err := txnMngr.AccountCurrencyBalance(fromUser.AccountID, txnArgs.CurrencyType, adapter)
	if err != nil {
		errMsg := fmt.Sprintf("failed to compute balance for account id %s and currency id %s, due to %s",
			fromUser.AccountID, txnArgs.CurrencyType, err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	var commission float32

	if txnArgs.CommissionType == SenderCommission {
		// Check whenever we have enough money to transfer and pay commission
		commission, err = txnMngr.CommissionForTransfer(txnArgs, adapter)
		if err != nil {
			// Failed to get commission, aborting
			logger.Error(err)
			return err
		}
	}
	// Commission is on sender hence need to check whenever he
	// has enough funds including commission,
	// is commissionType != SenderCommission => commission will be zero
	err = txnMngr.isEnoughBalance(currencyBalance, txnArgs.Amount+txnArgs.Amount*commission)

	if err != nil {
		errMsg := fmt.Sprintf("cannot lock funds for account id %s, due to %s", fromUser.AccountID, err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Update state by adding transaction which locks user funds
	outKey := fromUser.OutKeyForCurrency(txnArgs.CurrencyType)
	// We locking funds hence transfer to the lock account
	// TODO: Do we need to validate lock account? Whenever such account exits and there is system user associated with it.
	inKey := txnArgs.LockAccount.InKeyForCurrency(txnArgs.CurrencyType)

	// transaction record in a form:

	/*
		####################################################################################################################
			out.CurrencyType.FromUserAccountID,...,FromUserID,in.CurrencyType.LockAccountUserID (LockAccountUserID == system account id)
		####################################################################################################################
		 	need to add reciprocal transaction record, e.g.:
		 	in.CurrencyType.FromUserAccountID,...,SystemUserID,out.CurrencyType.FromUserAccountID (LockAccountUserID == system account id)
		####################################################################################################################
	*/

	// Add lock amount transaction
	logger.Debugf("Locking funds %f on account id %s", txnArgs.Amount, fromUser.AccountID)

	tt, err := adapter.GetTime()
	if err != nil {
		errMsg := fmt.Sprintf("failed to obtain tx time, due to %s", err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Copy transaction data
	if err := (&TransactionRecord{
		Key:           outKey,
		CurrencyType:  txnArgs.CurrencyType,
		User:          *fromUser,
		Amount:        txnArgs.Amount,
		OperationID:   txnArgs.OperationID,
		OperationType: LockAmount,
		ProcessType:   txnArgs.ProcessType,
		HashCode:      "", // TODO: Ask what is the meaning of the hash code here
		DependencyKey: inKey,
		Timestamp:     tt,
	}).Save(adapter); err != nil {
		return err
	}

	// Create reciprocal transaction with inverse keys and system user
	if err := (&TransactionRecord{
		Key:          inKey,
		CurrencyType: txnArgs.CurrencyType,
		User: UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:      "systemUserID",
			AccountInfo: txnArgs.LockAccount,
		},
		Amount:        txnArgs.Amount,
		OperationID:   txnArgs.OperationID,
		OperationType: LockAmount,
		ProcessType:   txnArgs.ProcessType,
		HashCode:      "", // TODO: Ask what is the meaning of the hash code here
		DependencyKey: outKey,
		Timestamp:     tt,
	}).Save(adapter); err != nil {
		return err
	}

	// If commission is on sender we need to lock commissions funds either
	if txnArgs.CommissionType == SenderCommission {
		// Copy transaction data
		commissionInKey := txnArgs.CommissionAccount.InKeyForCurrency(txnArgs.CurrencyType)
		logger.Debugf("Locking commission funds %f on account id %s", commission*txnArgs.Amount, fromUser.AccountID)
		if err := (&TransactionRecord{
			Key:           outKey,
			CurrencyType:  txnArgs.CurrencyType,
			User:          *fromUser,
			Amount:        commission * txnArgs.Amount,
			OperationID:   txnArgs.OperationID,
			OperationType: LockCommission,
			ProcessType:   txnArgs.ProcessType,
			HashCode:      "", // TODO: Ask what is the meaning of the hash code here
			DependencyKey: commissionInKey,
			Timestamp:     tt,
		}).Save(adapter); err != nil {
			return nil
		}

		// The reciprocal commission lock record
		if err := (&TransactionRecord{
			Key:          commissionInKey,
			CurrencyType: txnArgs.CurrencyType,
			User: UserAccountInfo{
				// Keep default values for time being, has to be replaced with
				// call to system chaincode to obtain system user and account information
				UserID:      "systemUserID",
				AccountInfo: txnArgs.CommissionAccount,
			},
			Amount:        commission * txnArgs.Amount,
			OperationID:   txnArgs.OperationID,
			OperationType: LockCommission,
			ProcessType:   txnArgs.ProcessType,
			HashCode:      "", // TODO: Ask what is the meaning of the hash code here
			DependencyKey: outKey,
			Timestamp:     tt,
		}).Save(adapter); err != nil {
			return nil
		}
	}

	// Funds locked successfully
	return nil
}

// isEnoughBalance check whenever there is enough balance
func (txnMngr *transactionManagerImpl) isEnoughBalance(balance, amount float32) error {
	if balance < amount {
		errMsg := "not enough funds to transfer money."
		logger.Error(errMsg)
		// We do not have enough money to transfer => returning the error
		return errors.New(errMsg)
	}
	return nil
}

// CommissionForTransfer compute commission based on transaction input parameters
func (txnMngr *transactionManagerImpl) CommissionForTransfer(txnArgs *TxnFundsLockInput, adapter *managerAux) (float32, error) {
	// TODO: Do we need to check correctness of the partner code?
	args := util.ToChaincodeArgs(GetCommissionFunction, txnArgs.PartnerCode, txnArgs.CustomerType, txnArgs.CommissionType.String())

	resp := adapter.InvokeChaincode(CommissionChaincode, args, CommissionChannel)
	if resp.Status != int32(shim.OK) {
		errMsg := fmt.Sprintf("Error obtain commission for operation id %s, parnter code %s, due to %s", txnArgs.OperationID, txnArgs.PartnerCode, resp.Message)
		logger.Error(errMsg)
		return 0, errors.New(errMsg)
	}

	var commission commissionCC.Commission
	err := json.Unmarshal(resp.Payload, &commission)
	if err != nil {
		errMsg := fmt.Sprintf("Failed to deserialize commission for operation id %s, parnter code %s, due to %s", txnArgs.OperationID, txnArgs.PartnerCode, resp.Message)
		logger.Error(errMsg)
		return 0, errors.New(errMsg)
	}
	cpercent, err := strconv.ParseFloat(commission.CommisionPercent, 32)
	if err != nil {
		errMsg := fmt.Sprintf("Failed to parse commission percent for operation id %s, parnter code %s, due to %s", txnArgs.OperationID, txnArgs.PartnerCode, resp.Message)
		logger.Error(errMsg)
		return 0, errors.New(errMsg)
	}
	return float32(cpercent), nil
}

// checkAndLoadUser check user account information and load in case it's only partially available
func (txnMngr *transactionManagerImpl) checkAndLoadUser(user *UserAccountInfo, adapter *managerAux) (*UserAccountInfo, error) {
	if user, err := txnMngr.accountVerifier(user, adapter); err != nil {
		errMsg := fmt.Sprintf("user account information is not valid (%s), error %s", user, err)
		logger.Errorf(errMsg)
		return nil, errors.New(errMsg)
	} else {
		// Account was either verified or loaded with missing information
		return user, nil
	}
}

// AccountCurrencyBalance compute user balance for given currency
func (txnMngr *transactionManagerImpl) AccountCurrencyBalance(accountID string, currencyType string, adapter *managerAux) (float32, error) {
	var total float32

	inKey := fmt.Sprintf("in.%s.%s", currencyType, accountID)
	outKey := fmt.Sprintf("out.%s.%s", currencyType, accountID)

	inQuery, err := json.Marshal(&querySelector{
		Selector: queryKey{
			Key: inKey,
		},
	})
	if err != nil {
		errMsg := fmt.Sprintf("Cannot serialize selector for all incoming transactions, due to %s", err)
		logger.Error(errMsg)
		return 0, errors.New(errMsg)
	}
	logger.Debugf("Incoming transactions selection query %s", inQuery)

	outQuery, err := json.Marshal(&querySelector{
		Selector: queryKey{
			Key: outKey,
		},
	})
	if err != nil {
		errMsg := fmt.Sprintf("cannot serialize selector for all outgoing transactions, due to %s", err)
		logger.Error(errMsg)
		return 0, errors.New(errMsg)
	}
	logger.Debugf("Outgoing transactions selection query %s", outQuery)

	// Compute total for all incoming transactions transactions
	it, err := adapter.GetQueryResult(string(inQuery))
	if err != nil {
		errMsg := fmt.Sprintf("failed to obtain results for incomming transactions, due to %s", err)
		logger.Error(errMsg)
		return 0, errors.New(errMsg)
	}

	// Iterate over all incoming transaction and add its amount to the total
	for it.HasNext() {
		txnRec, err := it.Next()
		if err != nil {
			errMsg := fmt.Sprintf("failed to get next transaction record, account id %s, due to %s", accountID, err)
			logger.Error(errMsg)
			return 0, errors.New(errMsg)
		}

		var tx TransactionRecord
		err = json.Unmarshal(txnRec.Value, &tx)
		if err != nil {
			errMsg := fmt.Sprintf("failed to unmarshal transaction record, for account id %s, due to %s", accountID, err)
			logger.Error(errMsg)
			return 0, errors.New(errMsg)
		}

		total += tx.Amount
	}

	// Compute total for all outgoing transactions
	it, err = adapter.GetQueryResult(string(outQuery))
	if err != nil {
		errMsg := fmt.Sprintf("failed to obtain results for outgoing transactions, due to %s", err)
		logger.Error(errMsg)
		return 0, errors.New(errMsg)
	}

	// Iterate over all outgoing transactions and subtract from the total
	for it.HasNext() {
		txnRec, err := it.Next()
		if err != nil {
			errMsg := fmt.Sprintf("failed to get next transaction record, account id %s, due to %s", accountID, err)
			logger.Error(errMsg)
			return 0, errors.New(errMsg)
		}

		var tx TransactionRecord
		err = json.Unmarshal(txnRec.Value, &tx)
		if err != nil {
			errMsg := fmt.Sprintf("failed to unmarshal transaction record, for account id %s, due to %s", accountID, err)
			logger.Error(errMsg)
			return 0, errors.New(errMsg)
		}

		total -= tx.Amount
	}

	return total, nil
}

const (
	LAST_LOAD_TIME = "LAST_LOAD_TIME"
	LAST_LOAD_KEYS = "LAST_LOAD_KEYS"
)

type currencyTypeKeys map[string]map[string]bool

type LastLoadKeys struct {
	MapCurrencyTypeKeys currencyTypeKeys `json:"keys"`
}

// запись информации о транзакции
func (txnMngr *transactionManagerImpl) Add(mapCurrencyTypeKeys currencyTypeKeys, currencyType string, transactionId string) currencyTypeKeys {
	if len(mapCurrencyTypeKeys) == 0 {
		mapCurrencyTypeKeys = make(currencyTypeKeys)
	}

	keys, ok := mapCurrencyTypeKeys[currencyType]
	if !ok {
		keys = make(map[string]bool)
		mapCurrencyTypeKeys[currencyType] = keys
	}
	keys[transactionId] = true

	return mapCurrencyTypeKeys
}

const (
	WAITING_SECONDS = 5 * 60
)

type CryptocurrencyResult struct {
	TransactionID string `json:"transaction_id"`
	Status        string `json:"status"`
	CommonMessage string `json:"common_message"`
	DetailMessage string `json:"detail_message"`
}

type ImportCryptocurrencyResult struct {
	CryptocurrencyResultArray []CryptocurrencyResult `json:"transaction_result"`
}

// loadCryptocurrency adds new cryptocurrency into transactionchannel
func (t *transactionManagerImpl) LoadCryptocurrency(cryptocurrencyTypeInput CryptocurrencyTypeInput, adapter *managerAux) (*ImportCryptocurrencyResult, error) {
	logger.Debug("CryptocurrencyChaincode::LoadCryptocurrency ENTER")
	defer logger.Debug("CryptocurrencyChaincode::LoadCryptocurrency EXIT")

	// проверка последнего времени страбатывания таймера
	invokeTime, err := adapter.GetTime()
	if err != nil {
		errMsg := fmt.Sprintf("Failed to GetTxTimestamp. %s", err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	// проверка времени отработки шедулера
	err = t.CheckLastLoadTime(invokeTime, adapter)
	if err != nil {
		return nil, err
	}

	// проверка что перед шедулером в канале transactionchannel сработал вызов в канале cryptocurrencychannel
	lastImportTime := t.getLastImportTime()
	if lastImportTime > invokeTime {
		errMsg := fmt.Sprintf("invokeTime dose not loaded")
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	// загрузка новых криптовалют из канала cryptocurrencychannel
	cryptocurrencyStoreOutput, err := t.cryptocurrencyLoader(cryptocurrencyTypeInput, adapter)

	// получить последние импортированные ключи
	lastLoadKeys, err := t.getLastLoadKeys(adapter)
	if err != nil {
		errMsg := fmt.Sprintf("Invalid CryptocurrencyListInput input format, failed due to %s", err)
		logger.Error(errMsg)
		return nil, errors.New(errMsg)
	}

	var importCryptocurrencyResult ImportCryptocurrencyResult
	for _, cryptocurrencyStore := range cryptocurrencyStoreOutput.CryptocurrencyStore {
		err := t.depositFundsTransaction(cryptocurrencyStore, adapter)
		if err != nil {
			result := CryptocurrencyResult{
				TransactionID: cryptocurrencyStore.TransactionID,
				Status:        "ERROR",
				CommonMessage: "Can't save transaction",
				DetailMessage: err.Error(),
			}
			importCryptocurrencyResult.CryptocurrencyResultArray = append(importCryptocurrencyResult.CryptocurrencyResultArray, result)
		} else {
			t.Add(lastLoadKeys.MapCurrencyTypeKeys, cryptocurrencyStore.CurrencyType, cryptocurrencyStore.TransactionID)
			result := CryptocurrencyResult{
				TransactionID: cryptocurrencyStore.TransactionID,
				Status:        "OK",
			}
			importCryptocurrencyResult.CryptocurrencyResultArray = append(importCryptocurrencyResult.CryptocurrencyResultArray, result)
		}
	}

	// запись последних импортированных ключейн
	err = t.SaveLastLoadKeys(lastLoadKeys, adapter)
	if err != nil {
		return nil, err
	}

	// сохранение времени срабатывания шедулера
	err = t.SaveLastLoadTime(invokeTime, adapter)
	if err != nil {
		return nil, err
	}

	return &importCryptocurrencyResult, nil
}

// сохранение времени отработки шедулера
func (t *transactionManagerImpl) SaveLastLoadTime(invokeTime int64, adapter *managerAux) error {
	invokeTimeByteArray := make([]byte, 8)
	binary.LittleEndian.PutUint64(invokeTimeByteArray, uint64(invokeTime))
	err := adapter.PutState(LAST_LOAD_TIME, invokeTimeByteArray)
	if err != nil {
		return err
	}

	return nil
}

// сохранить последние импортированные ключи
func (t *transactionManagerImpl) SaveLastLoadKeys(lastLoadKeys LastLoadKeys, adapter *managerAux) error {
	jsonLastLoadKeys, err := json.Marshal(lastLoadKeys)
	if err != nil {
		errMsg := fmt.Sprintf("failed to marhsal lastImportKeys, error %s", err)
		logger.Error(errMsg)
		return fmt.Errorf(errMsg)
	}

	err = adapter.PutState(LAST_LOAD_KEYS, jsonLastLoadKeys)
	if err != nil {
		errMsg := fmt.Sprintf("failed to PutState LAST_IMPORT_KEYS, error %s", err)
		logger.Error(errMsg)
		return fmt.Errorf(errMsg)
	}
	return nil
}

// получить список последних импортированных ключей
func (txnMngr *transactionManagerImpl) getLastLoadKeys(adapter *managerAux) (LastLoadKeys, error) {
	var lastLoadKeys LastLoadKeys
	lastImportKeysByteArray, err := adapter.GetState(LAST_LOAD_KEYS)
	if err != nil {
		errMsg := fmt.Sprintf("Failed to retreive LAST_IMPORT_KEYS. %s", err)
		logger.Error(errMsg)
		return lastLoadKeys, fmt.Errorf(errMsg)
	}

	if lastImportKeysByteArray == nil || len(lastImportKeysByteArray) == 0 {
		return lastLoadKeys, nil
	}

	err = json.Unmarshal(lastImportKeysByteArray, &lastLoadKeys)
	if err != nil {
		errMsg := fmt.Sprintf("Invalid lastLoadKeys input format, failed due to %s", err)
		logger.Error(errMsg)
		return lastLoadKeys, fmt.Errorf(errMsg)
	}

	return lastLoadKeys, nil
}

// сохранение транзакции на ввод криптовалюты
func (txnMngr *transactionManagerImpl) depositFundsTransaction(cryptocurrencyStore CryptocurrencyStore, adapter *managerAux) error {
	tt, err := adapter.GetTime()
	if err != nil {
		errMsg := fmt.Sprintf("failed to obtain tx time, due to %s", err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	inKey := cryptocurrencyStore.InAccountInfo.InKeyForCurrency(cryptocurrencyStore.CurrencyType)
	outKey := cryptocurrencyStore.InAccountInfo.InKeyForCurrency(cryptocurrencyStore.CurrencyType)

	// OUT
	if err := (&TransactionRecord{
		Key:          outKey,
		CurrencyType: cryptocurrencyStore.CurrencyType,
		User: UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:      "systemUserID",
			AccountInfo: AccountInfo{
				cryptocurrencyStore.OutAccountInfo.WalletID,
				cryptocurrencyStore.OutAccountInfo.AccountID,
			},
		},
		Amount:        cryptocurrencyStore.Amount,
		OperationID:   "",
		OperationType: TransferAmountToUser,
		ProcessType:   DepositFunds,
		HashCode:      "", // TODO: Ask what is the meaning of the hash code here
		DependencyKey: inKey,
		Timestamp:     tt,
	}).Save(adapter); err != nil {
		return err
	}

	// IN
	if err := (&TransactionRecord{
		Key:          inKey,
		CurrencyType: cryptocurrencyStore.CurrencyType,
		User: UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:      cryptocurrencyStore.InAccountInfo.UserID,
			AccountInfo: AccountInfo{
				cryptocurrencyStore.InAccountInfo.WalletID,
				cryptocurrencyStore.InAccountInfo.AccountID,
			},
		},
		Amount:        cryptocurrencyStore.Amount,
		OperationID:   "",
		OperationType: TransferAmountToUser,
		ProcessType:   DepositFunds,
		HashCode:      "", // TODO: Ask what is the meaning of the hash code here
		DependencyKey: outKey,
		Timestamp:     tt,
	}).Save(adapter); err != nil {
		return err
	}

	return nil
}

// проверка времени отработки шедулера
func (txnMngr *transactionManagerImpl) CheckLastLoadTime(currentInvokeTime int64, adapter *managerAux) error {
	lastLoadTimeByteArray, err := adapter.GetState(LAST_LOAD_TIME)
	if err != nil {
		return err
	}

	var lastLoadTime int64

	if lastLoadTimeByteArray == nil {
		lastLoadTime = currentInvokeTime - WAITING_SECONDS
	} else {
		lastLoadTime = int64(binary.LittleEndian.Uint64(lastLoadTimeByteArray))
	}

	if currentInvokeTime < lastLoadTime+WAITING_SECONDS {
		// считаем сколько нам осталось еще подождать
		needToWait := WAITING_SECONDS - (currentInvokeTime - lastLoadTime)
		needToWaitMsg := fmt.Sprintf("needToWait : %s", needToWait)
		logger.Info(needToWaitMsg)
		errMsg := fmt.Sprintf("You need to wait to %d seconds", needToWait)
		logger.Info(errMsg)
		return fmt.Errorf(errMsg)
	}

	return nil
}

type CryptocurrencyStore struct {
	Time           int64                      `json:"time"`
	CurrencyType   string                     `json:"currency_type"`
	TransactionID  string                     `json:"transaction_id"`
	OutAccountInfo *CryptocurrencyAccountInfo `json:"out_account_info_id"`
	InAccountInfo  *CryptocurrencyAccountInfo `json:"in_account_info_id"`
	Amount         float32                    `json:"amount"`
}

type CryptocurrencyAccountInfo struct {
	UserID           string `json:"user_id"`
	WalletID         string `json:"wallet_id"`
	AccountID        string `json:"account_id"`
	ExternalWalletID string `json:"external_wallet_id"`
}

type CryptocurrencyStoreOutput struct {
	CryptocurrencyStore []CryptocurrencyStore `json:"cryptocurrencies"`
}

func (txnMngr *transactionManagerImpl) getLastImportTime() int64 {
	return -1
}

func getLastLoadTime(currentInvokeTime int64, stub chaincodeStubInterfaceAdapter) (int64, error) {
	lastImportTimeByteArray, err := stub.GetState(LAST_LOAD_TIME)
	if err != nil {
		return -1, err
	}
	var lastImportTime int64

	if lastImportTimeByteArray == nil {
		lastImportTime = currentInvokeTime
	} else {
		lastImportTime = int64(binary.LittleEndian.Uint64(lastImportTimeByteArray))
	}

	return lastImportTime, nil
}

// TransferFunds finalize transferring funds from locked account to the end user
func (txnMngr *transactionManagerImpl) TransferFunds(txn *TxnFundsLockInput, adapter *managerAux) error {
	if len(txn.CurrencyType) == 0 {
		errMsg := fmt.Sprintf("currency id for operation %s is not provided", txn.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Verify from user information
	_, err := txnMngr.checkAndLoadUser(&txn.FromUser, adapter)
	if err != nil {
		logger.Error(err)
		return err
	}

	// Verify to user information
	_, err = txnMngr.checkAndLoadUser(&txn.ToUser, adapter)
	if err != nil {
		logger.Error(err)
		return err
	}

	outKey := txn.FromUser.OutKeyForCurrency(txn.CurrencyType)
	inKey := txn.LockAccount.InKeyForCurrency(txn.CurrencyType)

	// get locked funds transaction record
	lockedFundsRecord, err := querySelector{
		Selector: queryKey{
			Key:           outKey,
			DependencyKey: inKey,
			OperationID:   txn.OperationID,
		},
	}.selectTx(adapter)

	if err != nil {
		logger.Error(err.Error())
		return err
	}

	// Locked commission transaction record
	var lockedCommissionRecord *TransactionRecord
	// if commission is on sender need to check whenever transaction of
	// commission lock is exists
	if txn.CommissionType == SenderCommission {
		// In key for funds locked for commission
		inKey := txn.CommissionAccount.InKeyForCurrency(txn.CurrencyType)

		// get commission record from the ledger
		lockedCommissionRecord, err = querySelector{
			Selector: queryKey{
				Key:           outKey,
				DependencyKey: inKey,
				OperationID:   txn.OperationID,
			},
		}.selectTx(adapter)

		if err != nil {
			logger.Error(err.Error())
			return err
		}
	}

	// Check whenever funds on locked account was spent
	if err := txnMngr.isFundsSpent(txn.LockAccount.OutKeyForCurrency(txn.CurrencyType), txn.OperationID, adapter); err != nil {
		return err
	}

	// if commission is on sender need to check whenever commission was locked on temp account
	if txn.CommissionType == SenderCommission {
		// Check whenever funds on locked account for commission was spent
		if err := txnMngr.isFundsSpent(txn.CommissionAccount.OutKeyForCurrency(txn.CurrencyType), txn.OperationID, adapter); err != nil {
			return err
		}
	}

	outKey = txn.LockAccount.OutKeyForCurrency(txn.CurrencyType)
	// We locking funds hence transfer to the lock account
	inKey = txn.ToUser.InKeyForCurrency(txn.CurrencyType)

	var commission float32
	if txn.CommissionType == ReceiverCommission {
		// Compute commission, since its value was not committed during funds locking
		// transaction
		commission, err = txnMngr.CommissionForTransfer(txn, adapter)
		if err != nil {
			logger.Error(err.Error())
			return err
		}
	}

	// Compute amount for transfer based on commission
	// if commissionType is not ReceiverCommission then
	// commission will be zero hence amount will remain same
	systemPayment := commission * lockedFundsRecord.Amount
	amountForTransfer := lockedFundsRecord.Amount - systemPayment

	tt, err := adapter.GetTime()
	if err != nil {
		errMsg := fmt.Sprintf("failed to obtain tx time, due to %s", err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Copy transaction data
	logger.Debugf("Transferring funds from locked account %s to account id %s", txn.LockAccount.AccountID, txn.ToUser.AccountID)
	if err := (&TransactionRecord{
		Key:          outKey,
		CurrencyType: txn.CurrencyType,
		User: UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:      "systemUserID",
			AccountInfo: txn.LockAccount,
		},
		Amount:        amountForTransfer,
		OperationID:   txn.OperationID,
		OperationType: TransferAmountToUser,
		ProcessType:   txn.ProcessType,
		HashCode:      "", // TODO: Ask what is the meaning of the hash code here
		DependencyKey: inKey,
		Timestamp:     tt,
	}).Save(adapter); err != nil {
		return err
	}

	if err := (&TransactionRecord{
		Key:           inKey,
		CurrencyType:  txn.CurrencyType,
		User:          txn.ToUser,
		Amount:        amountForTransfer,
		OperationID:   txn.OperationID,
		OperationType: TransferAmountToUser,
		ProcessType:   txn.ProcessType,
		HashCode:      "", // TODO: Ask what is the meaning of the hash code here
		DependencyKey: outKey,
		Timestamp:     tt,
	}).Save(adapter); err != nil {
		return err
	}

	// 4. Transfer commission
	if txn.CommissionType == SenderCommission {
		systemPayment = lockedCommissionRecord.Amount

		outKey = txn.CommissionAccount.OutKeyForCurrency(txn.CurrencyType)
		// We locking funds hence transfer to the lock account
		//TODO: obtain system wallet and account ids
		logger.Warn("TODO! Update to obtain system account id")
		inKey = fmt.Sprintf("in.%s.systemAccountID", txn.CurrencyType)

		logger.Debugf("Transferring commission from locked account %s to the system account", txn.CommissionAccount.AccountID)
		if err := (&TransactionRecord{
			Key:          outKey,
			CurrencyType: txn.CurrencyType,
			User: UserAccountInfo{
				// Keep default values for time being, has to be replaced with
				// call to system chaincode to obtain system user and account information
				UserID:      "systemUserID",
				AccountInfo: txn.LockAccount,
			},
			Amount:        systemPayment,
			OperationID:   txn.OperationID,
			OperationType: TransferCommissionToSystem,
			ProcessType:   txn.ProcessType,
			HashCode:      "", // TODO: Ask what is the meaning of the hash code here
			DependencyKey: inKey,
			Timestamp:     tt,
		}).Save(adapter); err != nil {
			return nil
		}

		if err := (&TransactionRecord{
			Key:           inKey,
			CurrencyType:  txn.CurrencyType,
			User:          systemUser,
			Amount:        systemPayment,
			OperationID:   txn.OperationID,
			OperationType: TransferCommissionToSystem,
			ProcessType:   txn.ProcessType,
			HashCode:      "", // TODO: Ask what is the meaning of the hash code here
			DependencyKey: outKey,
			Timestamp:     tt,
		}).Save(adapter); err != nil {
			return nil
		}
	}

	return nil
}

// ReleaseFunds release funds which was locked into account
func (txnMngr *transactionManagerImpl) ReleaseFunds(txn *TxnFundsLockInput, adapter *managerAux) error {
	if len(txn.CurrencyType) == 0 {
		errMsg := fmt.Sprintf("currency id for operation %s is not provided", txn.OperationID)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Verify from user information
	_, err := txnMngr.checkAndLoadUser(&txn.FromUser, adapter)
	if err != nil {
		logger.Error(err)
		return err
	}

	// Verify to user information
	_, err = txnMngr.checkAndLoadUser(&txn.ToUser, adapter)
	if err != nil {
		logger.Error(err)
		return err
	}

	outKey := txn.FromUser.OutKeyForCurrency(txn.CurrencyType)
	inKey := txn.LockAccount.InKeyForCurrency(txn.CurrencyType)

	// Check whenever funds on locked account was spent
	if err := txnMngr.isFundsSpent(txn.LockAccount.OutKeyForCurrency(txn.CurrencyType), txn.OperationID, adapter); err != nil {
		return err
	}

	// get locked funds transaction record
	lockedFundsRecord, err := querySelector{
		Selector: queryKey{
			Key:           outKey,
			DependencyKey: inKey,
			OperationID:   txn.OperationID,
		},
	}.selectTx(adapter)

	if err != nil {
		logger.Error(err.Error())
		return err
	}

	tt, err := adapter.GetTime()
	if err != nil {
		errMsg := fmt.Sprintf("failed to obtain tx time, due to %s", err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	// Release funds by moving them back to the source account
	logger.Debugf("Releasing funds from locked account %s to account id %s", txn.LockAccount.AccountID, txn.FromUser.AccountID)
	if err := (&TransactionRecord{
		Key:          txn.LockAccount.OutKeyForCurrency(txn.CurrencyType),
		CurrencyType: txn.CurrencyType,
		User: UserAccountInfo{
			// Keep default values for time being, has to be replaced with
			// call to system chaincode to obtain system user and account information
			UserID:      "systemUserID",
			AccountInfo: txn.LockAccount,
		},
		Amount:        lockedFundsRecord.Amount,
		OperationID:   txn.OperationID,
		OperationType: TransferAmountToUser,
		ProcessType:   txn.ProcessType,
		HashCode:      "", // TODO: Ask what is the meaning of the hash code here
		DependencyKey: txn.FromUser.InKeyForCurrency(txn.CurrencyType),
		Timestamp:     tt,
	}).Save(adapter); err != nil {
		return nil
	}

	if err := (&TransactionRecord{
		Key:           txn.FromUser.InKeyForCurrency(txn.CurrencyType),
		CurrencyType:  txn.CurrencyType,
		User:          txn.FromUser,
		Amount:        lockedFundsRecord.Amount,
		OperationID:   txn.OperationID,
		OperationType: TransferAmountToUser,
		ProcessType:   txn.ProcessType,
		HashCode:      "", // TODO: Ask what is the meaning of the hash code here
		DependencyKey: txn.LockAccount.OutKeyForCurrency(txn.CurrencyType),
		Timestamp:     tt,
	}).Save(adapter); err != nil {
		return nil
	}

	if txn.CommissionType == SenderCommission {
		// Check if commission was already transferred
		if err := txnMngr.isFundsSpent(txn.CommissionAccount.OutKeyForCurrency(txn.CurrencyType), txn.OperationID, adapter); err != nil {
			return err
		}

		// In key for funds locked for commission
		inKey := txn.CommissionAccount.InKeyForCurrency(txn.CurrencyType)

		// get commission record from ledger
		lockedCommissionRecord, err := querySelector{
			Selector: queryKey{
				Key:           outKey,
				DependencyKey: inKey,
				OperationID:   txn.OperationID,
			},
		}.selectTx(adapter)

		if err != nil {
			logger.Error(err.Error())
			return err
		}

		logger.Debugf("Releasing commission from locked account %s back to the user account %s", txn.CommissionAccount.AccountID, txn.FromUser.AccountID)
		if err := (&TransactionRecord{
			Key:          txn.CommissionAccount.OutKeyForCurrency(txn.CurrencyType),
			CurrencyType: txn.CurrencyType,
			User: UserAccountInfo{
				// Keep default values for time being, has to be replaced with
				// call to system chaincode to obtain system user and account information
				UserID:      "systemUserID",
				AccountInfo: txn.CommissionAccount,
			},
			Amount:        lockedCommissionRecord.Amount,
			OperationID:   txn.OperationID,
			OperationType: TransferAmountToUser,
			ProcessType:   txn.ProcessType,
			HashCode:      "", // TODO: Ask what is the meaning of the hash code here
			DependencyKey: txn.FromUser.InKeyForCurrency(txn.CurrencyType),
			Timestamp:     tt,
		}).Save(adapter); err != nil {
			return nil
		}

		if err := (&TransactionRecord{
			Key:           txn.FromUser.InKeyForCurrency(txn.CurrencyType),
			CurrencyType:  txn.CurrencyType,
			User:          txn.FromUser,
			Amount:        lockedCommissionRecord.Amount,
			OperationID:   txn.OperationID,
			OperationType: TransferAmountToUser,
			ProcessType:   txn.ProcessType,
			HashCode:      "", // TODO: Ask what is the meaning of the hash code here
			DependencyKey: txn.CommissionAccount.OutKeyForCurrency(txn.CurrencyType),
			Timestamp:     tt,
		}).Save(adapter); err != nil {
			return nil
		}
	}

	return nil
}

func (txnMngr *transactionManagerImpl) isFundsSpent(key string, operationID string, adapter *managerAux) error {
	lockedSpentTx, err := json.Marshal(&querySelector{
		Selector: queryKey{
			Key:         key,
			OperationID: operationID,
		},
	})
	if err != nil {
		errMsg := fmt.Sprintf("failed to deserialize query selector to check locked account is unspent, due to %s", err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	it, err := adapter.GetQueryResult(string(lockedSpentTx))
	if err != nil {
		errMsg := fmt.Sprintf("failed to check whenever locked account was spent, due to %s", err)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	if it.HasNext() {
		// there is a transaction record with output related to the locked account
		errMsg := fmt.Sprintf("cannot spent smae transaction twice, locked funds has been used, the key %s", key)
		logger.Error(errMsg)
		return errors.New(errMsg)
	}

	return nil
}
