package txnmngr

import (
	"encoding/json"
	"fmt"
	"testing"

	"time"

	"github.com/hyperledger/fabric/common/util"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/hyperledger/fabric/protos/ledger/queryresult"
	"github.com/hyperledger/fabric/protos/peer"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/mock"
	"gitlab.smartblocklab.com/smartcontracts/common/commissionCC"
)

// ChaincodeStubInterfaceAdapterMock chaincode stub interface
type ChaincodeStubInterfaceAdapterMock struct {
	mock.Mock
}

func (stub *ChaincodeStubInterfaceAdapterMock) InvokeChaincode(chaincodeName string, args [][]byte, channel string) peer.Response {
	return stub.Called(chaincodeName, args, channel).Get(0).(peer.Response)
}

func (stub *ChaincodeStubInterfaceAdapterMock) GetTxID() string {
	return stub.Called().Get(0).(string)
}

func (stub *ChaincodeStubInterfaceAdapterMock) GetState(key string) ([]byte, error) {
	args := stub.Called(key)
	return args.Get(0).([]byte), args.Error(1)
}

func (stub *ChaincodeStubInterfaceAdapterMock) PutState(key string, value []byte) error {
	args := stub.Called(key, value)
	return args.Error(0)
}

func (stub *ChaincodeStubInterfaceAdapterMock) GetQueryResult(query string) (shim.StateQueryIteratorInterface, error) {
	args := stub.Called(query)
	if args.Get(0) == nil {
		return nil, nil
	}
	return args.Get(0).(shim.StateQueryIteratorInterface), args.Error(1)
}

// stateQueryIteratorInterfaceMock query result iterator
type stateQueryIteratorInterfaceMock struct {
	mock.Mock
}

func (query *stateQueryIteratorInterfaceMock) Close() error {
	return nil
}

func (query *stateQueryIteratorInterfaceMock) Next() (*queryresult.KV, error) {
	return query.Called().Get(0).(*queryresult.KV), query.Called().Error(1)
}

func (query *stateQueryIteratorInterfaceMock) HasNext() bool {
	return query.Called().Get(0).(bool)
}

func TestTransactionManagerImplLockFundsOnAccountValidAccountHavingEnoughFunds(t *testing.T) {
	txnMngr := NewTransactionManager(func(user *UserAccountInfo, adapter chaincodeStubInterfaceAdapter) (*UserAccountInfo, error) {
		return user, nil
	}, func(cryptocurrencyTypeInput CryptocurrencyTypeInput, adapter chaincodeStubInterfaceAdapter) (*CryptocurrencyStoreOutput, error) {
		return nil, nil
	})

	adapter := &ChaincodeStubInterfaceAdapterMock{}

	adapter.On("GetTxID").Return("testTXID")

	input := &TxnFundsLockInput{
		CurrencyType:     "101",
		Amount:         100,
		ProcessType:    ExchangeWithCommission,
		OperationID:    "1",
		PartnerCode:    "testPartnerCode",
		CustomerType:   "testCustomerType",
		CommissionType: SenderCommission,
		FromUser: UserAccountInfo{
			UserID: "fromUserID",
			AccountInfo: AccountInfo{
				WalletID:  "fromWalletID",
				AccountID: "fromAccountID",
			},
		},
		ToUser: UserAccountInfo{
			UserID: "toUserID",
			AccountInfo: AccountInfo{
				WalletID:  "toWalletID",
				AccountID: "toAccountID",
			},
		},
		LockAccount: AccountInfo{
			AccountID: "lockAccountID",
			WalletID:  "lockWalletID",
		},
		CommissionAccount: AccountInfo{
			AccountID: "commissionAccountID",
			WalletID:  "commissionWalletID",
		},
	}

	incomingQueryIterator := &stateQueryIteratorInterfaceMock{}
	incomingQueryIterator.On("HasNext").
		Return(true).
		Once().
		On("HasNext").
		Return(false)

	inRecordValue, err := json.Marshal(&TransactionRecord{
		Key:    "in.101.fromAccountID",
		Amount: 100,
	})
	assert.NoError(t, err)

	incomingQueryIterator.On("Next").
		Return(&queryresult.KV{
			Namespace: "ns1",
			Key:       "in.101.fromAccountID",
			Value:     inRecordValue,
		}, nil)

	outgoingQueryIterator := &stateQueryIteratorInterfaceMock{}
	outgoingQueryIterator.On("HasNext").Return(false)
	adapter.On("GetQueryResult", "{\"selector\":{\"key\":\"in.101.fromAccountID\"}}").
		Return(incomingQueryIterator, nil).
		Once().
		On("GetQueryResult", "{\"selector\":{\"key\":\"out.101.fromAccountID\"}}").
		Return(outgoingQueryIterator, nil)

	adapter.On("PutState", fmt.Sprintf("testTXID_fromUserID_%d", LockAmount), mock.Anything).Run(func(params mock.Arguments) {
		// Check lock funds transaction write state input
		jsonRecord := params.Get(1).([]byte)
		var record TransactionRecord
		err := json.Unmarshal(jsonRecord, &record)
		// Should be able to unmarshal w/o error
		assert.NoError(t, err)
		assert.Equal(t, "out.101.fromAccountID", record.Key)
		assert.Equal(t, "in.101.lockAccountID", record.DependencyKey)
		assert.Equal(t, float32(100), record.Amount)
		assert.Equal(t, "101", record.CurrencyType)
		assert.Equal(t, "fromUserID", record.User.UserID)
		assert.Equal(t, "fromAccountID", record.User.AccountID)
		assert.Equal(t, "fromWalletID", record.User.WalletID)
	}).
		Return(nil).
		Once().
		On("PutState", fmt.Sprintf("testTXID_fromUserID_%d", LockCommission), mock.Anything).Run(func(params mock.Arguments) {
		// Check transaction lock commission funds record
		jsonRecord := params.Get(1).([]byte)
		var record TransactionRecord
		err := json.Unmarshal(jsonRecord, &record)
		// Should be able to unmarshal w/o error
		assert.NoError(t, err)
		assert.Equal(t, "out.101.fromAccountID", record.Key)
		assert.Equal(t, "in.101.commissionAccountID", record.DependencyKey)
		assert.Equal(t, float32(0), record.Amount)
		assert.Equal(t, "101", record.CurrencyType)
		assert.Equal(t, "fromUserID", record.User.UserID)
		assert.Equal(t, "fromAccountID", record.User.AccountID)
		assert.Equal(t, "fromWalletID", record.User.WalletID)
	}).
		Return(nil).
		Once().
		On("PutState", fmt.Sprintf("testTXID_systemUserID_%d", LockAmount), mock.Anything).Run(func(params mock.Arguments) {
		// Check lock funds transaction write state input
		jsonRecord := params.Get(1).([]byte)
		var record TransactionRecord
		err := json.Unmarshal(jsonRecord, &record)
		// Should be able to unmarshal w/o error
		assert.NoError(t, err)
		assert.Equal(t, "in.101.lockAccountID", record.Key)
		assert.Equal(t, "out.101.fromAccountID", record.DependencyKey)
		assert.Equal(t, float32(100), record.Amount)
		assert.Equal(t, "101", record.CurrencyType)
		assert.Equal(t, "systemUserID", record.User.UserID)
		assert.Equal(t, "lockAccountID", record.User.AccountID)
		assert.Equal(t, "lockWalletID", record.User.WalletID)
	}).
		Return(nil).
		Once().
		On("PutState", fmt.Sprintf("testTXID_systemUserID_%d", LockCommission), mock.Anything).Run(func(params mock.Arguments) {
		// Check transaction lock commission funds record
		jsonRecord := params.Get(1).([]byte)
		var record TransactionRecord
		err := json.Unmarshal(jsonRecord, &record)
		// Should be able to unmarshal w/o error
		assert.NoError(t, err)
		assert.Equal(t, "in.101.commissionAccountID", record.Key)
		assert.Equal(t, "out.101.fromAccountID", record.DependencyKey)
		assert.Equal(t, float32(0), record.Amount)
		assert.Equal(t, "101", record.CurrencyType)
		assert.Equal(t, "systemUserID", record.User.UserID)
		assert.Equal(t, "commissionAccountID", record.User.AccountID)
		assert.Equal(t, "commissionWalletID", record.User.WalletID)
	}).
		Return(nil)

	args :=
		util.ToChaincodeArgs(GetCommissionFunction, input.PartnerCode, input.CustomerType, string(input.CommissionType))
	commissionJson, err := json.Marshal(commissionCC.Commission{
		CommisionPercent: "0",
	})
	assert.NoError(t, err)
	adapter.On("InvokeChaincode", CommissionChaincode, args, CommissionChannel).Return(shim.Success(commissionJson))

	err = txnMngr.LockFundsOnAccount(input, &managerAux{
		chaincodeStubInterfaceAdapter: adapter,
		timeGetterFunc: func() (int64, error) {
			return int64(time.Now().Second()), nil
		},
	})
	// We do not expect any error
	assert.NoError(t, err)
}

func TestTransactionManagerImplLockFundsOnAccountValidAccountHavingNotEnoughFunds(t *testing.T) {
	txnMngr := NewTransactionManager(func(user *UserAccountInfo, adapter chaincodeStubInterfaceAdapter) (*UserAccountInfo, error) {
		return user, nil
	}, func(cryptocurrencyTypeInput CryptocurrencyTypeInput, adapter chaincodeStubInterfaceAdapter) (*CryptocurrencyStoreOutput, error) {
		return nil, nil
	})

	adapter := &ChaincodeStubInterfaceAdapterMock{}
	adapter.On("GetTxID").Return("testTXID")

	// Only necessary information for current test case
	input := &TxnFundsLockInput{
		CurrencyType:     "101",
		Amount:         100,
		ProcessType:    ExchangeWithCommission,
		OperationID:    "1",
		PartnerCode:    "testPartnerCode",
		CustomerType:   "testCustomerType",
		CommissionType: SenderCommission,
		FromUser: UserAccountInfo{
			UserID: "fromUserID",
			AccountInfo: AccountInfo{
				WalletID:  "fromWalletID",
				AccountID: "fromAccountID",
			},
		},
		ToUser: UserAccountInfo{
			UserID: "toUserID",
		},
	}

	incomingQueryIterator := &stateQueryIteratorInterfaceMock{}
	incomingQueryIterator.On("HasNext").
		Return(true).
		Once().
		On("HasNext").
		Return(false)

	inRecordValue, err := json.Marshal(&TransactionRecord{
		Key:    "in.101.fromAccountID",
		Amount: 150,
	})
	assert.NoError(t, err)

	incomingQueryIterator.On("Next").
		Return(&queryresult.KV{
			Namespace: "ns1",
			Key:       "in.101.fromAccountID",
			Value:     inRecordValue,
		}, nil)

	outgoingQueryIterator := &stateQueryIteratorInterfaceMock{}
	outgoingQueryIterator.On("HasNext").
		Return(true).
		Once().
		On("HasNext").
		Return(false)

	outRecordValue, err := json.Marshal(&TransactionRecord{
		Key:    "out.101.fromAccountID",
		Amount: 100,
	})
	assert.NoError(t, err)

	outgoingQueryIterator.On("Next").
		Return(&queryresult.KV{
			Namespace: "ns1",
			Key:       "in.101.fromAccountID",
			Value:     outRecordValue,
		}, nil)

	adapter.On("GetQueryResult", "{\"selector\":{\"key\":\"in.101.fromAccountID\"}}").
		Return(incomingQueryIterator, nil).
		Once().
		On("GetQueryResult", "{\"selector\":{\"key\":\"out.101.fromAccountID\"}}").
		Return(outgoingQueryIterator, nil)

	args := util.ToChaincodeArgs(GetCommissionFunction, input.PartnerCode, input.CustomerType, string(input.CommissionType))
	commissionJson, err := json.Marshal(commissionCC.Commission{
		CommisionPercent: "0",
	})
	assert.NoError(t, err)
	adapter.On("InvokeChaincode", CommissionChaincode, args, CommissionChannel).Return(shim.Success(commissionJson))

	err = txnMngr.LockFundsOnAccount(input, &managerAux{
		chaincodeStubInterfaceAdapter: adapter,
		timeGetterFunc: func() (int64, error) {
			return int64(time.Now().Second()), nil
		},
	})
	// We do not have enough funds resulting with error
	assert.Error(t, err)
	// Expecting error which will tell there is not enough funds
	assert.Contains(t, err.Error(), "cannot lock funds")
}

func TestTransactionManagerImplLockFundsOnAccountValidAccountHavingEnoughFundsNoCommission(t *testing.T) {
	txnMngr := NewTransactionManager(func(user *UserAccountInfo, adapter chaincodeStubInterfaceAdapter) (*UserAccountInfo, error) {
		return user, nil
	}, func(cryptocurrencyTypeInput CryptocurrencyTypeInput, adapter chaincodeStubInterfaceAdapter) (*CryptocurrencyStoreOutput, error) {
		return nil, nil
	})

	adapter := &ChaincodeStubInterfaceAdapterMock{}
	adapter.On("GetTxID").Return("testTXID")

	input := &TxnFundsLockInput{
		CurrencyType:     "101",
		Amount:         100,
		ProcessType:    ExchangeWithCommission,
		OperationID:    "1",
		PartnerCode:    "testPartnerCode",
		CustomerType:   "testCustomerType",
		CommissionType: ReceiverCommission,
		FromUser: UserAccountInfo{
			UserID: "fromUserID",
			AccountInfo: AccountInfo{
				WalletID:  "fromWalletID",
				AccountID: "fromAccountID",
			},
		},
		ToUser: UserAccountInfo{
			UserID: "toUserID",
			AccountInfo: AccountInfo{
				WalletID:  "toWalletID",
				AccountID: "toAccountID",
			},
		},
		LockAccount: AccountInfo{
			AccountID: "lockAccountID",
			WalletID:  "lockWalletID",
		},
		CommissionAccount: AccountInfo{
			AccountID: "commissionAccountID",
			WalletID:  "commissionWalletID",
		},
	}

	incomingQueryIterator := &stateQueryIteratorInterfaceMock{}
	incomingQueryIterator.On("HasNext").
		Return(true).
		Once().
		On("HasNext").
		Return(false)

	inRecordValue, err := json.Marshal(&TransactionRecord{
		Key:    "in.101.fromAccountID",
		Amount: 100,
	})
	assert.NoError(t, err)

	incomingQueryIterator.On("Next").
		Return(&queryresult.KV{
			Namespace: "ns1",
			Key:       "in.101.fromAccountID",
			Value:     inRecordValue,
		}, nil)

	outgoingQueryIterator := &stateQueryIteratorInterfaceMock{}
	outgoingQueryIterator.On("HasNext").Return(false)
	adapter.On("GetQueryResult", "{\"selector\":{\"key\":\"in.101.fromAccountID\"}}").
		Return(incomingQueryIterator, nil).
		Once().
		On("GetQueryResult", "{\"selector\":{\"key\":\"out.101.fromAccountID\"}}").
		Return(outgoingQueryIterator, nil)

	adapter.On("PutState", fmt.Sprintf("testTXID_fromUserID_%d", LockAmount), mock.Anything).Run(func(params mock.Arguments) {
		// Check lock funds transaction write state input
		jsonRecord := params.Get(1).([]byte)
		var record TransactionRecord
		err := json.Unmarshal(jsonRecord, &record)
		// Should be able to unmarshal w/o error
		assert.NoError(t, err)
		assert.Equal(t, "out.101.fromAccountID", record.Key)
		assert.Equal(t, "in.101.lockAccountID", record.DependencyKey)
		assert.Equal(t, float32(100), record.Amount)
		assert.Equal(t, "101", record.CurrencyType)
		assert.Equal(t, "fromUserID", record.User.UserID)
		assert.Equal(t, "fromAccountID", record.User.AccountID)
		assert.Equal(t, "fromWalletID", record.User.WalletID)
	}).
		Return(nil).
		Once().
		On("PutState", fmt.Sprintf("testTXID_fromUserID_%d", LockCommission), mock.Anything).Run(func(params mock.Arguments) {
		// There should not be a call for adding commissions lock record
		t.Fail()

	}).Return(nil)
	adapter.On("PutState", fmt.Sprintf("testTXID_systemUserID_%d", LockAmount), mock.Anything).Run(func(params mock.Arguments) {
		// Check lock funds transaction write state input
		jsonRecord := params.Get(1).([]byte)
		var record TransactionRecord
		err := json.Unmarshal(jsonRecord, &record)
		// Should be able to unmarshal w/o error
		assert.NoError(t, err)
		assert.Equal(t, "in.101.lockAccountID", record.Key)
		assert.Equal(t, "out.101.fromAccountID", record.DependencyKey)
		assert.Equal(t, float32(100), record.Amount)
		assert.Equal(t, "101", record.CurrencyType)
		assert.Equal(t, "systemUserID", record.User.UserID)
		assert.Equal(t, "lockAccountID", record.User.AccountID)
		assert.Equal(t, "lockWalletID", record.User.WalletID)
	}).
		Return(nil).
		Once().
		On("PutState", fmt.Sprintf("testTXID_%d", LockCommission), mock.Anything).Run(func(params mock.Arguments) {
		// There should not be a call for adding commissions lock record
		t.Fail()

	}).Return(nil)

	args := util.ToChaincodeArgs(GetCommissionFunction, input.PartnerCode, input.CustomerType, string(input.CommissionType))
	commissionJson, err := json.Marshal(commissionCC.Commission{
		CommisionPercent: "0",
	})
	assert.NoError(t, err)
	adapter.On("InvokeChaincode", CommissionChaincode, args, CommissionChannel).Return(shim.Success(commissionJson))

	err = txnMngr.LockFundsOnAccount(input, &managerAux{
		chaincodeStubInterfaceAdapter: adapter,
		timeGetterFunc: func() (int64, error) {
			return int64(time.Now().Second()), nil
		},
	})
	// We do not expect any error
	assert.NoError(t, err)
}
